import argparse
import json

from rpft import converters
from rpft.logger.logger import initialize_main_logger


def main():
    initialize_main_logger()
    args = create_parser().parse_args()
    args.func(args)


def create_flows(args):
    flows = converters.create_flows(
        args.input,
        None,
        args.format,
        data_models=args.datamodels,
        tags=args.tags,
    )

    with open(args.output, "w", encoding="utf-8") as export:
        json.dump(flows, export, indent=4)


def convert_to_json(args):
    content = converters.convert_to_json(args.input, args.format)

    with open(args.output, "wb") as export:
        export.write(bytes(content, "utf-8"))


def flows_to_sheets(args):
    converters.flows_to_sheets(
        args.input, args.output, args.format, args.strip_uuids, args.numbered
    )


def uni_to_sheets(args):
    with open(args.output, "wb") as handle:
        handle.write(converters.uni_to_sheets(args.input))


def sheets_to_uni(args):
    data = converters.sheets_to_uni(args.input)

    with open(args.output, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2)


def create_parser():
    parser = argparse.ArgumentParser(
        description=("create RapidPro flows JSON from spreadsheets"),
    )
    sub = parser.add_subparsers(
        help="run {subcommand} --help for further information",
        required=True,
        title="subcommands",
    )

    _add_create_command(sub)
    _add_convert_command(sub)
    _add_flows_to_sheets_command(sub)
    _add_uni_to_sheets_command(sub)
    _add_sheets_to_uni_command(sub)

    return parser


def _add_create_command(sub):
    parser = sub.add_parser(
        "create",
        aliases=["create_flows"],
        help="create RapidPro flows from spreadsheets",
    )

    parser.set_defaults(func=create_flows)
    parser.add_argument(
        "input",
        help=(
            "paths to XLSX or JSON files, or directories containing CSV files, or"
            " Google Sheets IDs i.e. from the URL; inputs should be of the same format"
        ),
        nargs="+",
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output JSON filename",
        required=True,
    )
    parser.add_argument(
        "--tags",
        help=(
            "tags to filter the content index sheet; a sequence of lists, with each"
            " list starting with an integer (tag position) followed by tags to include"
            " or exclude for this position; exclusions are preceded by '!'; for"
            " example, the value '1 foo bar ! 2 !baz' includes rows where tags.1"
            " is 'foo', 'bar' or not empty, and tags:2 is not 'baz' or empty"
        ),
        nargs="*",
    )
    parser.add_argument(
        "--datamodels",
        help=(
            "name of the module defining user data models underlying the data sheets,"
            " e.g. if the model definitions reside in"
            " ./myfolder/mysubfolder/mymodelsfile.py, then this argument should be"
            " myfolder.mysubfolder.mymodelsfile"
        ),
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["csv", "google_sheets", "json", "uni", "xlsx"],
        help="input sheet format",
        required=True,
    )


def _add_convert_command(sub):
    parser = sub.add_parser("convert", help="save input spreadsheets as JSON")

    parser.set_defaults(func=convert_to_json)
    parser.add_argument(
        "-f",
        "--format",
        choices=["csv", "google_sheets", "json", "xlsx"],
        help="input sheet format",
        required=True,
    )
    parser.add_argument(
        "input",
        help=(
            "path to XLSX or JSON file, or directory containing CSV files, or Google"
            " Sheets ID i.e. from the URL"
        ),
    )
    parser.add_argument(
        "output",
        help=("path to output JSON file"),
    )


def _add_flows_to_sheets_command(sub):
    parser = sub.add_parser(
        "flows_to_sheets", help="convert RapidPro JSON into spreadsheets"
    )

    parser.set_defaults(func=flows_to_sheets)
    parser.add_argument(
        "--strip_uuids",
        action="store_true",
        help="strip all UUIDs from output to allow for comparing outputs",
    )
    parser.add_argument(
        "--numbered",
        action="store_true",
        help="Use sequential numbers instead of short representations for row IDs",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["csv", "xlsx"],
        help="desired sheet format (default: csv)",
        default="csv",
    )
    parser.add_argument(
        "input",
        help=("path to input RapidPro JSON file"),
    )
    parser.add_argument(
        "output",
        help=("output folder"),
    )


def _add_uni_to_sheets_command(sub):
    parser = sub.add_parser(
        "uni-to-sheets",
        help="convert JSON to sheets",
    )
    parser.set_defaults(func=uni_to_sheets)
    parser.add_argument(
        "input",
        help=("location of input JSON file"),
    )
    parser.add_argument(
        "output",
        help=("location where sheets will be saved"),
    )


def _add_sheets_to_uni_command(sub):
    parser = sub.add_parser(
        "sheets-to-uni",
        help="convert sheets to nested JSON",
    )

    parser.set_defaults(func=sheets_to_uni)
    parser.add_argument(
        "input",
        help=("location of workbook"),
    )
    parser.add_argument(
        "output",
        help=("location where JSON will be saved"),
    )


if __name__ == "__main__":
    main()



import logging

from rpft.rapidpro.models.common import Exit
from rpft.rapidpro.models.exceptions import RapidProRouterError
from rpft.rapidpro.utils import generate_new_uuid
from rpft.parsers.creation.flowrowmodel import Condition, Edge

logger = logging.getLogger(__name__)


class BaseRouter:
    def __init__(self, result_name=None, categories=None):
        self.type = None
        self.categories = categories or []
        self.result_name = result_name

    def from_dict(data, exits):
        if data["type"] == "random":
            return RandomRouter.from_dict(data, exits)
        elif data["type"] == "switch":
            return SwitchRouter.from_dict(data, exits)
        else:
            raise RapidProRouterError("Router data has invalid router type.")

    def _get_result_name_and_categories_from_data(data, exits):
        categories = [
            RouterCategory.from_dict(category_data, exits)
            for category_data in data["categories"]
        ]
        result_name = None
        if "result_name" in data:
            result_name = data["result_name"]
        return result_name, categories

    def set_result_name(self, result_name):
        # TODO: Check that this works
        self.result_name = result_name

    def _get_category_or_none(self, category_name):
        result = [c for c in self.get_categories() if c.name == category_name]
        if result:
            return result[0]

    def get_category_by_uuid(self, uuid):
        # It might be better if cases had a reference to
        # a category, rather than their uuid.
        result = [c for c in self.get_categories() if c.uuid == uuid]
        if result:
            return result[0]
        else:
            raise KeyError("No Category with given uuid")

    def _add_category(self, category_name, destination_uuid):
        category = RouterCategory(category_name, destination_uuid)
        self.categories.append(category)
        return self.categories[-1]

    def get_or_create_category(self, category_name, destination_uuid):
        # If the category exists, this updates its destination uuid
        category = self._get_category_or_none(category_name)
        if category:
            category.update_destination_uuid(destination_uuid)
            return category
        else:
            return self._add_category(category_name, destination_uuid)

    def get_exits(self):
        return [c.get_exit() for c in self.get_categories()]

    def record_global_uuids(self, uuid_dict):
        pass

    def assign_global_uuids(self, uuid_dict):
        pass

    def get_categories(self):
        # TODO: Remove this and have the default category included in the list for the
        # SwitchRouter
        raise NotImplementedError

    def render(self):
        raise NotImplementedError

    def get_exit_edge_pairs(self, row_id):
        raise NotImplementedError


class SwitchRouter(BaseRouter):
    def __init__(
        self,
        operand,
        result_name=None,
        wait_timeout=None,
        cases=None,
        categories=None,
        default_category=None,
        no_response_category=None,
    ):
        """
        wait_timeout:
            None: There is no waiting for a user's message in this router
            0: Waiting for user's message, without a timeout
            >0: Waiting for user's message, with a timeout (in which case
                no_response_category is taken)
        """

        super().__init__(result_name)
        self.type = "switch"
        self.operand = operand
        self.cases = cases or []
        self.wait_timeout = wait_timeout

        self.has_explicit_default_category = False

        if default_category:
            self.default_category = default_category
            # Indicates that a default category has been added by the user
            self.has_explicit_default_category = True
        else:
            # Add an implicit default category
            category = RouterCategory("Other", None)
            self.default_category = category
        self.categories = categories or []

        self.has_explicit_no_response_category = False
        if self.wait_timeout:
            if no_response_category:
                self.no_response_category = no_response_category
                # Indicates that a No Response category has been added by the user
                self.has_explicit_no_response_category = True
            else:
                category = RouterCategory("No Response", None)
                self.no_response_category = category
        else:
            self.no_response_category = None
            if no_response_category:
                logger.warning(
                    "Router has No Response category but no wait timeout."
                    " Ignoring No Response category."
                )

    def from_dict(data, exits):
        result_name, categories = BaseRouter._get_result_name_and_categories_from_data(
            data, exits
        )
        cases = [RouterCase.from_dict(case_data) for case_data in data["cases"]]
        default_categories = [
            category
            for category in categories
            if category.uuid == data["default_category_uuid"]
        ]
        if not default_categories:
            raise RapidProRouterError(
                "Default category uuid does not match any category."
            )
        no_response_category = None
        no_response_category_id = None
        wait_timeout = None
        if "wait" in data:
            wait_timeout = 0
            if "timeout" in data["wait"]:
                no_response_category_id = data["wait"]["timeout"]["category_uuid"]
                wait_timeout = int(data["wait"]["timeout"]["seconds"])
                no_response_categories = [
                    category
                    for category in categories
                    if category.uuid == no_response_category_id
                ]
                if not no_response_categories:
                    raise RapidProRouterError(
                        "No Response category uuid does not match any category."
                    )
                no_response_category = no_response_categories[0]
        other_categories = [
            category
            for category in categories
            if category.uuid
            not in [data["default_category_uuid"], no_response_category_id]
        ]
        return SwitchRouter(
            data["operand"],
            result_name,
            wait_timeout,
            cases,
            other_categories,
            default_categories[0],
            no_response_category,
        )

    def _get_case_or_none(self, comparison_type, arguments):
        for case in self.cases:
            if case.type == comparison_type and case.arguments == arguments:
                return case

    def _add_case(self, comparison_type, arguments, category_uuid):
        case = RouterCase(comparison_type, arguments, category_uuid)
        self.cases.append(case)
        return self.cases[-1]

    def create_case(self, comparison_type, arguments, category):
        return self._add_case(comparison_type, arguments, category.uuid)

    def generate_category_name(self, comparison_arguments):
        # Auto-generate a category name that is guaranteed to be unique
        category_name = "_".join([str(a).title() for a in comparison_arguments])
        category_name = category_name or "Yes"
        while self._get_category_or_none(category_name):
            category_name += "_alt"
        return category_name

    def add_choice(
        self,
        comparison_variable,
        comparison_type,
        comparison_arguments,
        category_name,
        destination_uuid,
        is_default=False,
    ):
        self.set_operand(comparison_variable)
        case = self._get_case_or_none(comparison_type, comparison_arguments)
        if case:
            # Case already exists. Only update the destination
            # category_name is ignored.
            category = self.get_category_by_uuid(case.category_uuid)
            category.update_destination_uuid(destination_uuid)
            return

        if not category_name:
            category_name = self.generate_category_name(comparison_arguments)
        if is_default:
            self.update_default_category(destination_uuid, category_name)
            category = self.default_category
        else:
            category = self.get_or_create_category(category_name, destination_uuid)
        self._add_case(comparison_type, comparison_arguments, category.uuid)
        return category

    def update_default_category(self, destination_uuid, category_name=None):
        if self.has_explicit_default_category:
            logger.warning("Overwriting default category for Router")
        self.default_category.update_destination_uuid(destination_uuid)
        if category_name:
            self.default_category.update_name(category_name)
        if destination_uuid or category_name:
            self.has_explicit_default_category = True

    def update_no_response_category(self, destination_uuid, category_name=None):
        if not self.wait_timeout:
            logger.warning(
                "Updating No Response category, but router has no wait timeout."
            )
        if self.has_explicit_no_response_category:
            logger.warning("Overwriting No Response category for Router")
        self.no_response_category.update_destination_uuid(destination_uuid)
        if category_name:
            self.no_response_category.update_name(category_name)
        if destination_uuid or category_name:
            self.has_explicit_no_response_category = True

    def has_wait(self):
        return self.wait_timeout is not None

    def has_positive_wait(self):
        return bool(self.wait_timeout)

    def get_categories(self):
        if self.no_response_category:
            return (
                self.categories + [self.default_category] + [self.no_response_category]
            )
        else:
            return self.categories + [self.default_category]

    def set_operand(self, operand):
        if not operand:
            return
        if self.operand and operand and self.operand != operand:
            logger.warning(f"Overwriting operand from {self.operand} -> {operand}")

        self.operand = operand

    def record_global_uuids(self, uuid_dict):
        for case in self.cases:
            if case.type == "has_group":
                uuid_dict.record_group_uuid(case.arguments[1], case.arguments[0])

    def assign_global_uuids(self, uuid_dict):
        for case in self.cases:
            if case.type == "has_group":
                case.arguments[0] = uuid_dict.get_group_uuid(case.arguments[1])

    def validate(self):
        # TODO: Add more validation
        if self.wait_timeout is not None:
            assert type(self.wait_timeout) is int
            assert self.wait_timeout >= 0
            if self.wait_timeout > 0:
                assert self.no_response_category is not None

    def render(self):
        self.validate()
        render_dict = {
            "type": self.type,
            "operand": self.operand,
            "cases": [case.render() for case in self.cases],
            "categories": [category.render() for category in self.get_categories()],
            "default_category_uuid": self.default_category.uuid,
        }
        if self.has_positive_wait():
            render_dict.update(
                {
                    "wait": {
                        "type": "msg",
                        "timeout": {
                            "seconds": self.wait_timeout,
                            "category_uuid": self.no_response_category.uuid,
                        },
                    }
                }
            )
        elif self.has_wait():
            render_dict.update(
                {
                    "wait": {
                        "type": "msg",
                    }
                }
            )
        if self.result_name is not None:
            render_dict.update({"result_name": self.result_name})
        return render_dict

    def get_exit_edge_pairs(self, row_id):
        pairs = []
        covered_category_uuids = set()

        for category in self.get_categories():
            for case_ in self.cases:
                if case_.category_uuid == category.uuid:
                    covered_category_uuids.add(category.uuid)

                    try:
                        arg = case_.arguments[
                            1 if case_.type == "has_group" else 0
                        ]
                    except IndexError:
                        arg = None

                    if self.operand in ["@contact.groups", "@child.run.status"]:
                        # For groups and expired/complete, var/type/name are implicit
                        condition = Condition(value=arg or "")
                    else:
                        value = (
                            arg
                            if case_.type not in RouterCase.NO_ARGS_TESTS
                            and arg is not None
                            else ""
                        )
                        condition = Condition(
                            value=value,
                            variable=self.operand,
                            type=case_.type,
                            name=category.name,
                        )
                    pairs.append(
                        (category.exit, Edge(from_=row_id, condition=condition))
                    )
                    break

        if self.default_category.uuid not in covered_category_uuids:
            # Sometimes, the default category is already covered by a specific case.
            # In that case, its edge has already been covered.
            # By convention, the condition is blank rather than 'Other'
            pairs.append((self.default_category.exit, Edge(from_=row_id)))

        if self.no_response_category:
            pairs.append(
                (
                    self.no_response_category.exit,
                    Edge(from_=row_id, condition=Condition(value=category.name)),
                )
            )

        return pairs


class RandomRouter(BaseRouter):
    def __init__(self, result_name=None, categories=None):
        super().__init__(result_name, categories)
        self.type = "random"

    def from_dict(data, exits):
        result_name, categories = BaseRouter._get_result_name_and_categories_from_data(
            data, exits
        )
        return RandomRouter(result_name, categories)

    def add_choice(self, category_name=None, destination_uuid=None):
        self.get_or_create_category(
            category_name if category_name else f"Bucket {len(self.categories) + 2}",
            destination_uuid,
        )

    def get_categories(self):
        return self.categories

    def validate(self):
        # TODO: Add validation
        pass

    def render(self):
        render_dict = {
            "type": self.type,
            "categories": [category.render() for category in self.categories],
        }
        if self.result_name:
            render_dict.update({"result_name": self.result_name})
        return render_dict

    def get_exit_edge_pairs(self, row_id):
        pairs = []
        for category in self.categories:
            pairs.append(
                (
                    category.exit,
                    Edge(from_=row_id, condition=Condition(value=category.name)),
                )
            )
        return pairs


class RouterCategory:
    def __init__(
        self, name, destination_uuid=None, uuid=None, exit_uuid=None, exit=None
    ):
        """
        :param name: Name of the category
        :param destination_uuid: The UUID of the node that this category should point to

        The destination of the category can be provided either as a destination_uuid or
        an exit.
        """
        self.uuid = uuid or generate_new_uuid()
        if len(name) > 115:
            raise RapidProRouterError(f"Category name too long (>115): {name}")
        self.name = name
        if exit:
            self.exit = exit
        else:
            self.exit = Exit(
                uuid=exit_uuid or generate_new_uuid(), destination_uuid=destination_uuid
            )

    def from_dict(data, exits):
        """
        :param data: The router in json format, as in RapidPro flows
        :param exits: A list of exit objects containing an exit the category connects to
        """
        matching_exits = [exit for exit in exits if exit.uuid == data["exit_uuid"]]
        if not matching_exits:
            raise RapidProRouterError("RouterCategory with no matching exit.")
        return RouterCategory(
            name=data["name"], uuid=data["uuid"], exit=matching_exits[0]
        )

    def get_exit(self):
        return self.exit

    def update_destination_uuid(self, uuid):
        self.exit.destination_uuid = uuid

    def update_name(self, name):
        self.name = name

    def render(self):
        return {
            "uuid": self.uuid,
            "name": self.name,
            "exit_uuid": self.exit.uuid,
        }


class RouterCase:
    NO_ARGS_TESTS = {
        "has_date",
        "has_email",
        "has_error",
        "has_number",
        "has_state",
        "has_text",
        "has_time",
    }

    TEST_VALIDATIONS = {
        "all_words": lambda x: len(x) == 1 and x[0],
        "has_any_word": lambda x: len(x) == 1 and x[0],
        "has_beginning": lambda x: len(x) == 1 and x[0],
        "has_category": lambda x: len(x) >= 1 and all(x),
        "has_date": lambda x: len(x) == 0,
        "has_date_eq": lambda x: len(x) == 1 and x[0],
        "has_date_gt": lambda x: len(x) == 1 and x[0],
        "has_date_lt": lambda x: len(x) == 1 and x[0],
        "has_district": lambda x: len(x) == 1 and x[0],
        "has_email": lambda x: len(x) == 0,
        "has_error": lambda x: len(x) == 0,
        # For has_group: First is UUID and second is name.
        # In imported flows, UUID is obligatory, but the toolkit
        # uses blank as a placeholder. Instead, the toolkit requires
        # a name, which usually is optional.
        "has_group": lambda x: len(x) in {1, 2} and (x[0] or x[1]),
        "has_intent": lambda x: len(x) == 2 and all(x),
        "has_number": lambda x: len(x) == 0,
        "has_number_between": lambda x: len(x) == 2 and all(x),
        "has_number_eq": lambda x: len(x) == 1 and x[0],
        "has_number_gt": lambda x: len(x) == 1 and x[0],
        "has_number_gte": lambda x: len(x) == 1 and x[0],
        "has_number_lt": lambda x: len(x) == 1 and x[0],
        "has_number_lte": lambda x: len(x) == 1 and x[0],
        "has_only_phrase": lambda x: len(x) == 1 and x[0],
        "has_only_text": lambda x: len(x) == 1 and x[0],
        "has_pattern": lambda x: len(x) == 1 and x[0],
        "has_phone": lambda x: len(x) in {0, 1} and all(x),
        "has_phrase": lambda x: len(x) == 1 and x[0],
        "has_state": lambda x: len(x) == 0,
        "has_text": lambda x: len(x) == 0,
        "has_time": lambda x: len(x) == 0,
        "has_top_intent": lambda x: len(x) == 2 and all(x),
        "has_ward": lambda x: len(x) == 2 and all(x),
    }

    def __init__(self, comparison_type, arguments, category_uuid, uuid=None):
        self.uuid = uuid or generate_new_uuid()
        self.type = comparison_type
        self.category_uuid = category_uuid
        if self.type in RouterCase.NO_ARGS_TESTS:
            self.arguments = []
        else:
            self.arguments = arguments
        self.validate()

    def from_dict(data):
        return RouterCase(
            data["type"], data["arguments"], data["category_uuid"], data["uuid"]
        )

    def validate(self):
        if self.type not in RouterCase.TEST_VALIDATIONS:
            raise RapidProRouterError(f'Invalid router test type: "{self.type}"')
        if not RouterCase.TEST_VALIDATIONS[self.type](self.arguments):
            raise RapidProRouterError(
                f"Invalid number of arguments {len(self.arguments)} or blank "
                f'arguments for router test type "{self.type}"'
            )

    def render(self):
        self.validate()
        return {
            "uuid": self.uuid,
            "type": self.type,
            "category_uuid": self.category_uuid,
            "arguments": self.arguments,
        }



import json
import logging
from collections import ChainMap
from logging.config import dictConfig
from pathlib import Path

from rpft.logger import DEFAULT_CONFIG


logger = logging.getLogger(__name__)


class LoggingContextHandler:
    def __init__(self):
        self.context_variables = []
        self.processing_stack = []

    def add(self, processing_unit, **new_context_vars):
        self.processing_stack.append(processing_unit)
        self.context_variables.append(new_context_vars)

    def get_processing_stack(self):
        return self.processing_stack

    def get_context_variables(self):
        # Union of all the dicts
        return dict(ChainMap(*self.context_variables))

    def pop(self):
        self.processing_stack.pop()
        self.context_variables.pop()


_context = LoggingContextHandler()


class logging_context:
    def __init__(self, processing_unit, **kwargs):
        self.processing_unit = processing_unit
        self.kwargs = kwargs

    def __enter__(self):
        _context.add(self.processing_unit, **self.kwargs)

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type is None:
            _context.pop()


class ContextFilter(logging.Filter):

    def filter(self, record):
        record.processing_stack = " | ".join(_context.get_processing_stack())
        record.context_variables = _context.get_context_variables()
        return True


def initialize_main_logger(file_path="errors.log", config_path="logging.json"):
    config = None

    if Path(config_path).exists():
        with open(config_path, "r") as f:
            config = json.load(f)
    else:
        config = dict(DEFAULT_CONFIG)

    config["handlers"]["file"]["filename"] = file_path
    dictConfig(config)
    logger.debug(f"Logging configured, config={config}")



from pydantic import model_validator

from rpft.parsers.common.rowparser import ParserModel
from rpft.parsers.creation.models import Condition


class Webhook(ParserModel):
    # Headers should ideally be a dict, once that is supported
    url: str = ""
    method: str = ""
    headers: list = []
    body: str = ""


def list_of_pairs_to_dict(headers):
    # Dict is not yet supported in the row parser,
    # so we need to convert a list of pairs into dict.
    if type(headers) is dict:
        return headers

    if type(headers) is list:
        if headers == [""]:
            return {}

        if not all(map(lambda x: type(x) is list and len(x) == 2, headers)):
            raise ValueError("Value must be a list of pairs.")

        return {k: v for k, v in headers}

    raise ValueError("Value must be a dict or list of pairs.")


def dict_to_list_of_pairs(headers):
    if type(headers) is list:
        return headers

    if type(headers) is dict:
        return [[k, v] for k, v in headers.items()]

    raise ValueError("Value must be a list/dict.")


class WhatsAppTemplating(ParserModel):
    name: str = ""
    uuid: str = ""
    variables: list[str] = []


class Edge(ParserModel):
    from_: str = ""
    condition: Condition = Condition()

    def header_name_to_field_name(header):
        field_map = {
            "from": "from_",
        }
        return field_map.get(header, header)

    def field_name_to_header_name(field):
        field_map = {
            "from_": "from",
        }
        return field_map.get(field, field)

    def header_name_to_field_name_with_context(header, row):
        return Edge.header_name_to_field_name(header)


class FlowRowModel(ParserModel):
    row_id: str = ""
    type: str
    edges: list[Edge]
    loop_variable: list[str] = []
    include_if: bool = True
    mainarg_message_text: str = ""
    mainarg_value: str = ""
    mainarg_groups: list[str] = []
    mainarg_none: str = ""
    mainarg_dict: list = []  # encoded as list of pairs
    mainarg_destination_row_ids: list[str] = []
    mainarg_flow_name: str = ""
    mainarg_expression: str = ""
    mainarg_iterlist: list = []
    wa_template: WhatsAppTemplating = WhatsAppTemplating()
    webhook: Webhook = Webhook()
    data_sheet: str = ""
    data_row_id: str = ""
    template_arguments: list = []
    choices: list[str] = []
    save_name: str = ""
    result_category: str = ""
    image: str = ""
    audio: str = ""
    video: str = ""
    attachments: list[str] = []
    urn_scheme: str = ""
    obj_name: str = ""
    obj_id: str = ""
    node_name: str = ""
    node_uuid: str = ""
    no_response: str = ""
    ui_type: str = ""
    ui_position: list[str] = []

    @model_validator(mode="before")
    def set_main_arg(cls, data):
        try:
            name = cls.header_name_to_field_name_with_context("message_text", data)
            data[name] = data.pop("message_text")
        except Exception:
            pass

        return data

    def field_name_to_header_name(field):
        field_map = {
            "node_uuid": "_nodeId",
            "ui_type": "_ui_type",
            "ui_position": "_ui_position",
            "mainarg_message_text": "message_text",
            "mainarg_value": "message_text",
            "mainarg_groups": "message_text",
            "mainarg_none": "message_text",
            "mainarg_destination_row_ids": "message_text",
            "mainarg_flow_name": "message_text",
            "mainarg_expression": "message_text",
            "mainarg_dict": "message_text",
            "webhook.body": "message_text",
        }
        return field_map.get(field, field)

    def header_name_to_field_name_with_context(header, row):
        basic_header_dict = {
            "from": "edges.*.from_",
            "condition": "edges.*.condition.value",
            "condition_value": "edges.*.condition.value",
            "condition_var": "edges.*.condition.variable",
            "condition_variable": "edges.*.condition.variable",
            "condition_type": "edges.*.condition.type",
            "condition_name": "edges.*.condition.name",
            "_nodeId": "node_uuid",
            "_ui_type": "ui_type",
            "_ui_position": "ui_position",
        }
        row_type_to_main_arg = {
            "send_message": "mainarg_message_text",
            "save_value": "mainarg_value",
            "add_to_group": "mainarg_groups",
            "remove_from_group": "mainarg_groups",
            "save_flow_result": "mainarg_value",
            "wait_for_response": "mainarg_none",
            "add_contact_urn": "mainarg_value",
            "set_contact_language": "mainarg_value",
            "set_contact_name": "mainarg_value",
            "set_contact_status": "mainarg_value",
            "set_contact_timezone": "mainarg_value",
            "split_random": "mainarg_none",
            "go_to": "mainarg_destination_row_ids",
            "call_webhook": "webhook.body",
            "transfer_airtime": "mainarg_dict",
            "start_new_flow": "mainarg_flow_name",
            "split_by_value": "mainarg_expression",
            "split_by_group": "mainarg_groups",
            "insert_as_block": "mainarg_flow_name",
            "begin_for": "mainarg_iterlist",
            "end_for": "mainarg_none",
            "begin_block": "mainarg_none",
            "end_block": "mainarg_none",
            "hard_exit": "mainarg_none",
            "loose_exit": "mainarg_none",
            "no_op": "mainarg_none",
        }

        if header in basic_header_dict:
            return basic_header_dict[header]

        if header == "message_text":
            return row_type_to_main_arg[row["type"]]

        return header

    def is_starting_row(self):
        return len(self.edges) == 1 and self.edges[0].from_ == "start"



import re
from collections import defaultdict
from collections.abc import Iterable, Sequence

from typing import List

from pydantic import BaseModel, ConfigDict, field_validator, model_validator

from rpft.parsers.common.cellparser import CellParser


def is_pairs(value):
    return all(
        isinstance(item, Sequence) and not isinstance(item, str) and len(item) == 2
        for item in value
    )


class RowParserError(Exception):
    pass


class ParserModel(BaseModel):

    model_config = ConfigDict(coerce_numbers_to_str=True)

    def header_name_to_field_name(header):
        # Given a human-friendly column header name, map it to the
        # string defining which field(s) in the model the cell
        # value corresponds to.
        # This is necessary as we might want
        # to use some string representations as fields which are
        # reserved words in python.
        return header

    def field_name_to_header_name(field):
        # Reverse of the above.
        # Currently, however, this is more limited, as field
        # only refers to top-level field entries, rather than
        # full hierarchies pointing to a field.
        return field

    def header_name_to_field_name_with_context(header, row):
        # This is used for models representing a full sheet row.
        return header

    @field_validator("*", mode="before")
    @classmethod
    def collect(cls, v, info):
        """
        Collect a single value into a list if the target field is a list
        """
        field = cls.model_fields[info.field_name]

        if is_list_type(field.annotation) and (
            not isinstance(v, Sequence) or isinstance(v, str)
        ):
            return [v] if v != "" else []

        return v

    @model_validator(mode="before")
    @classmethod
    def coerce_to_dict(cls, data):
        """
        Coerce different types of data into a dict, if possible.
        """
        if isinstance(data, str):
            data = [data]

        if isinstance(data, Sequence):
            if is_pairs(data):
                return dict(data)
            else:
                return {
                    name: val[1] if isinstance(val, list) and len(val) > 1 else val
                    for (name, _), val in zip(cls.model_fields.items(), data)
                }

        return data


def get_list_child_model(model):
    if is_basic_list_type(model):
        # If not specified, list elements may be anything.
        # Without additional information, we assume strings.
        child_model = str
    else:
        # Get the type that's inside the list
        assert len(model.__args__) == 1
        child_model = model.__args__[0]
    return child_model


def is_list_type(model):
    """
    Determine whether model is a list type, such as list, list[str], List, List[str].

    typing.List is deprecated as of Python 3.9
    """
    return (
        is_basic_list_type(model)
        or model is List
        or getattr(model, "__origin__", None) is list
    )


def is_basic_list_type(model):
    return model is list


def is_basic_dict_type(model):
    return model is dict


def is_list_instance(value):
    return isinstance(value, list)


def is_iterable_instance(value):
    return isinstance(value, Iterable) and not type(value) is str


def is_parser_model_type(model):
    # Determine whether model is a subclass of ParserModel.
    try:
        return issubclass(model, ParserModel)
    except TypeError:
        # This occurs in Python >= 3.7 if one argument is a nested type, e.g. List[str]
        return False


def is_parser_model_instance(value):
    return isinstance(value, ParserModel)


def is_basic_type(model):
    return model in (str, int, float, bool)


def is_basic_instance(value):
    return isinstance(value, (str, int, float, bool))


def is_default_value(model_instance, field, field_value):
    return field_value == type(model_instance).model_fields[field].default


def str_to_bool(string):
    # in this case, the default value takes effect.
    if string.lower() == "false":
        return False
    else:
        return True


def get_field_name(string):
    return (
        string.split(RowParser.TYPE_ANNOTATION_SEPARATOR)[0]
        .split(RowParser.DEFAULT_VALUE_SEPARATOR)[0]
        .strip()
    )


class RowParser:
    # Takes a dictionary of cell entries, whose keys are the column names
    # and the values are the cell content converted into nested lists.
    # Turns this into an instance of the provided model.

    HEADER_FIELD_SEPARATOR = "."
    TYPE_ANNOTATION_SEPARATOR = ":"
    DEFAULT_VALUE_SEPARATOR = "="

    def __init__(self, model, cell_parser=None):
        self.model = model
        self.output = None  # Gets reinitialized with each call to parse_row
        self.cell_parser = cell_parser or CellParser()

    def try_assign_as_kwarg(self, field, key, value, model):
        # If value can be interpreted as a (field, field_value) pair for a field of
        # model, assign value to field[key] (which represents the field in the model)
        if is_list_instance(value) and len(value) == 2 and type(value[0]) is str:
            first_entry_as_key = model.header_name_to_field_name(value[0])
            if first_entry_as_key in model.model_fields:
                self.assign_value(
                    field[key],
                    first_entry_as_key,
                    value[1],
                    model.model_fields[first_entry_as_key].annotation,
                )
                return True
        return False

    def assign_value(self, field, key, value, model):
        # Given a field in the output and a key, assign
        # value (which can be a nested structure) to field[key].
        # This is done recursively, and
        # converts some lists into dicts in the process,
        # as appropriate and as determined by model,
        # which is the model/type of field[key]).
        # Note: field can also be a list and key an index.

        # Using both key and field here because if we passed field[key],
        # we can't do call by reference with basic types.
        if is_parser_model_type(model):
            # The value should be a dict/object
            field[key] = {}
            # Get the list of keys that are available for the target model
            # Note: The fields have a well defined ordering.
            # See https://pydantic-docs.helpmanual.io/usage/models/#field-ordering
            model_fields = list(model.model_fields.keys())

            if type(value) is not list:
                # It could be that an object is specified via a single element.
                value = [value]
            if self.try_assign_as_kwarg(field, key, value, model):
                # Check if value could encode a single KWArg; if yes, assign as such.
                # Note: We're resolving an ambiguity here in favor of kwargs.
                # in principle, this could also be two positional arguments.
                return

            for i, entry in enumerate(value):
                # Go through the list of arguments
                kwarg_found = False
                if self.try_assign_as_kwarg(field, key, entry, model):
                    # This looks like a KWArg
                    # Note: We're resolving an ambiguity here in favor of kwargs.
                    # in principle, this could also be a positional argument that is a
                    # list.
                    kwarg_found = True
                else:
                    # This isn't a KWarg, so we interpret is as a positional argument
                    # KWArgs should come after positional arguments --> assert
                    assert not kwarg_found
                    entry_key = model_fields[i]
                    self.assign_value(
                        field[key],
                        entry_key,
                        entry,
                        model.model_fields[entry_key].annotation,
                    )
        elif is_basic_dict_type(model):
            field[key] = {}
            if not value:
                return
            if not is_iterable_instance(value):
                raise ValueError("dict-type cell must contain key-value pairs")
            value = list(value)
            if isinstance(value[0], str):
                assert len(value) == 2
                field[key] = {value[0]: value[1]}
            elif isinstance(value[0], list):
                for entry in value:
                    assert len(entry) == 2
                    field[key][entry[0]] = entry[1]
        elif is_basic_list_type(model):
            # We cannot iterate deeper if we don't know what to expect.
            if is_iterable_instance(value):
                field[key] = list(value)
            else:
                field[key] = [value]
        elif is_list_type(model):
            child_model = get_list_child_model(model)
            # The created entry should be a list. Value should also be a list
            field[key] = []
            # Note: This makes a decision on how to resolve an ambiguity when the target
            # field is a list of lists, but the cell value is a 1-dimensional list.
            # 1;2 → [[1],[2]] rather than [[1,2]]
            if not is_list_instance(value):
                # It could be that a list is specified via a single element.
                if value == "":
                    # Interpret an empty cell as [] rather than ['']
                    value = []
                else:
                    value = [value]
            for entry in value:
                # For each entry, create a new list entry and assign its value
                # recursively
                field[key].append(None)
                self.assign_value(field[key], -1, entry, child_model)
        else:
            assert is_basic_type(model)
            # The value should be a basic type
            # TODO: Ensure the types match. E.g. we don't want value to be a list
            if model == bool:
                if type(value) is str:
                    stripped = value.strip()
                    # Special case: empty string is not assigned at all.
                    if stripped:
                        field[key] = str_to_bool(stripped)
                else:
                    field[key] = bool(value)
            else:
                field[key] = model(value)

    def find_entry(self, model, output_field, field_path):
        # Within the output_field (which may be a nested structure),
        # traverse the field_path to find the subfield to assign the value to.
        # Return that field (via a parent object and a key, so that we can
        # overwrite its value) and its model.

        # Note: model is the model/type that the output_field should correspond to
        # (though objects are modeled as dicts in the output). It helps us
        # traverse the path in output_field and if necessary create non-existent
        # entries.

        # We're creating the output object's fields as we're going through it.
        # It'd be nicer to already have a template.
        field_name = field_path[0]
        if is_list_type(model):
            child_model = get_list_child_model(model)
            index = int(field_name) - 1
            if len(output_field) <= index:
                # Create a new list entry for this, if necessary
                # We assume the columns are always in order 1, 2, 3, ... for now
                assert len(output_field) == index
                # None will later be overwritten by assign_value
                output_field.append(None)

            key = index
        elif is_basic_dict_type(model):
            key = field_name
            output_field[key] = None
            child_model = str
        else:
            assert is_parser_model_type(model)
            key = model.header_name_to_field_name(field_name)
            if key not in model.model_fields:
                raise ValueError(f"Field {key} doesn't exist in target type {model}.")
            child_model = model.model_fields[key].annotation

            if key not in output_field:
                # Create a new entry for this, if necessary
                # None will later be overwritten by assign_value
                output_field[key] = None

        if len(field_path) == 1:
            # We've reached the end of the field_path
            # Therefore we've found where we need to assign
            return output_field, key, child_model
        else:
            # The field has subfields, keep going and recurse.
            # If field doesn't exist yet in our output object, create it.
            if is_list_type(child_model) and output_field[key] is None:
                output_field[key] = []
            elif is_basic_dict_type(child_model) and output_field[key] is None:
                output_field[key] = {}
            elif is_parser_model_type(child_model) and output_field[key] is None:
                output_field[key] = {}
            # recurse
            return self.find_entry(child_model, output_field[key], field_path[1:])

    def parse_entry(
        self, column_name, value, value_is_parsed=False, template_context={}
    ):
        # This creates/populates a field in self.output
        # The field is determined by column_name, its value by value
        column_name = get_field_name(column_name)
        field_path = column_name.split(RowParser.HEADER_FIELD_SEPARATOR)
        # Find the destination subfield in self.output that corresponds to field_path
        field, key, model = self.find_entry(self.model, self.output, field_path)
        # The destination field in self.output is field[key], its type is model.
        # Therefore the value should be assigned to field[key].
        # (Note: This is a bit awkward; if we returned field[key] itself, we could
        # not easily overwrite its value. So we return field and key separately.
        # Ideally we would return a pointer to the destination field.
        # The model of field[key] is model, and thus value should also be interpreted
        # as being of type model.
        if not value_is_parsed:
            if (
                is_list_type(model)
                or is_basic_dict_type(model)
                or is_parser_model_type(model)
            ):
                # If the expected type of the value is list/object,
                # parse the cell content as such.
                # Otherwise leave it as a string
                value = self.cell_parser.parse(value, context=template_context)
            else:
                value, _ = self.cell_parser.parse_as_string(
                    value, context=template_context
                )
        self.assign_value(field, key, value, model)

    def parse_row(self, data, template_context={}):
        # data is a dict where the keys are column header names,
        # and the values are the corresponding values of the cells
        # in the spreadsheet (i.e. strings).
        # However, because we don't have the string parser yet,
        # the values are assumed to be parsed already, i.e. are
        # nested lists.

        # Initialize the output template as a dict
        self.output = {}

        # Apply map from header string to field specification
        data_rekeyed = {}
        for k, v in data.items():
            k = self.model.header_name_to_field_name_with_context(k, data)
            data_rekeyed[k] = v
        data = data_rekeyed

        # For each column with an asterisk (*) (indicating list of fields),
        # Compute how long the implied list is by taking the maximum
        # over the lengths of all fields that this list refers to.
        # Note: So far, no nested asterisks are supported.
        asterisk_list_lengths = defaultdict(lambda: 1)
        for k, v in data.items():
            if "*" in k:
                prefix = k.split("*")[0]
                parsed_v = self.cell_parser.parse(v, context=template_context)
                if isinstance(parsed_v, list):
                    asterisk_list_lengths[prefix] = max(
                        asterisk_list_lengths[prefix], len(parsed_v)
                    )
                    # No else case needed because then the implied list length is 1,
                    # i.e. the default value
        # Process each entry
        for k, v in data.items():
            if "*" in k:
                # Process each prefix:*:suffix column entry by assigning the individual
                # list values to prefix:1:suffix, prefix:2:suffix, etc
                prefix = k.split("*")[0]
                parsed_v = self.cell_parser.parse(v, context=template_context)
                if not isinstance(parsed_v, list):
                    # If there was only one entry, we assume it is used for the entire
                    # list
                    parsed_v = [parsed_v] * asterisk_list_lengths[prefix]
                for i, elem in enumerate(parsed_v):
                    self.parse_entry(
                        k.replace("*", str(i + 1)),
                        elem,
                        value_is_parsed=True,
                        template_context=template_context,
                    )
            else:
                # Normal, non-* column entry.
                self.parse_entry(k, v, template_context=template_context)
        # Returning an instance of the model rather than the output directly
        # helps us fill in default values where no entries exist.
        # Filtering out None values here is a bit of a hack;
        # the cause of these is the line output_field[key] = None in find_key.
        # Ideally, we should fix the cause rather than clean up here.
        self.output = {k: v for k, v in self.output.items() if v is not None}
        return self.model(**self.output)

    def unparse_row(self, model_instance, target_headers=set(), excluded_headers=set()):
        """
        Turn a model instance into spreadsheet row.

        Args:
            model_instance (ParserModel): model instance to convert
            target_headers (set[str]): Complex type fields (ParserModels, lists, dicts)
                whose content should be represented in the output dict as a single
                entry. A trailing asterisk may be used to specify multiple fields
                at once, such as `list.*` and `field.*`.
            excluded_headers (set[str]): Fields to exclude from the output. Same format
                as target_headers.

        Returns:
            A flat dict[str,str] where the keys reference fields of the model
            and the values string representations of the values of the fields.
            Keys can also reference subfields, for example, if the model has a
            field `field` whose type is another model which has a field `subfield`,
            then `field.subfield` would be a valid key referencing this subfield.
            Similarly, `list.1` can reference to the first element of a list.
            By default, all model content is unravelled until the referenced fields
            are basic types. However, `target_headers` can be used to specify
            complex type fields (ParserModels, lists, dicts) whose content should be
            represented as a single string.
        """
        self.output_dict = {}
        self.unparse_row_recurse(model_instance, "", target_headers, excluded_headers)
        return self.output_dict

    def trim_prefix(self, prefix):
        # We have to remove the leading '.' from the prefix, if necessary
        if prefix.startswith("."):
            return prefix[1:]
        return prefix

    def write_to_output_dict(self, prefix, value):
        if not is_basic_instance(value):
            value_list = self.to_nested_list(value)
            value = self.cell_parser.join_from_lists(value_list)
        prefix = self.trim_prefix(prefix)
        if prefix in self.output_dict:
            old_value = self.output_dict[prefix]
            raise RowParserError(
                f'Unparse: Multiple entries ("{old_value}" and "{value}") '
                f'with same key "{prefix}."'
            )
        self.output_dict[prefix] = value

    def unparse_row_recurse(
        self, value, prefix, target_headers=set(), excluded_headers=set()
    ):
        if value is None or self.matches_headers(prefix, excluded_headers):
            return

        if is_basic_instance(value) or self.matches_headers(prefix, target_headers):
            self.write_to_output_dict(prefix, value)
        elif is_list_instance(value):
            for i, entry in enumerate(value):
                self.unparse_row_recurse(
                    entry,
                    f"{prefix}{RowParser.HEADER_FIELD_SEPARATOR}{i+1}",
                    target_headers,
                    excluded_headers,
                )
        elif is_parser_model_instance(value):
            for field, field_value in value:
                if is_default_value(value, field, field_value):
                    continue
                mapped_field = type(value).field_name_to_header_name(field)
                field_prefix = (
                    f"{prefix}{RowParser.HEADER_FIELD_SEPARATOR}{mapped_field}"
                )
                if field == mapped_field:
                    # No remapping happened
                    self.unparse_row_recurse(
                        field_value,
                        field_prefix,
                        target_headers,
                        excluded_headers,
                    )
                else:
                    # If a remapping occurs, we allow no further recursion.
                    # We would get inconsistencies where e.g. if we map a list
                    # and a string to the same key `mapped_field`, the string
                    # might generate a column `mapped_field` while the list may
                    # generated columns `mapped_field.1` and `mapped_field.2`.
                    if not self.matches_headers(field_prefix, excluded_headers):
                        self.write_to_output_dict(field_prefix, field_value)
        else:
            raise ValueError(f"Unsupported field type {type(value)} of {value}.")

    def matches_headers(self, prefix, target_headers):
        if not prefix:
            return False
        prefix = self.trim_prefix(prefix)
        # Examples:
        #     a.b.field matches a.b.field
        #     list.1 matches list.1
        #     list.1 matches list.*
        #     list.1.field matches list.*.field
        #     dict.field matches dict.*
        # Technically, x.field.subfield matches x.field; however, such a comparison
        # will never occur in unparse_row_recurse because once x.field is encountered,
        # the recursion bottoms out (because x.field matches x.field) and does not
        # proceed to process the x.field.subfield prefix.
        for header in target_headers:
            header_regex = "^" + header.replace(".", "\\.").replace("*", "[^.]+")
            pattern = re.compile(header_regex)
            if pattern.match(prefix):
                return True
        return False

    def to_nested_list(self, value):
        if is_basic_instance(value):
            return value
        elif is_list_instance(value):
            return [self.to_nested_list(e) for e in value]
        elif is_parser_model_instance(value):
            # We're encoding key-value pairs here, which takes a nesting depth of 2.
            # We could also consider encoding as positional arguments, however,
            # this is not reversible in the case where there are exactly two values,
            # and first value coincides with the name of a model attribute.
            value_dict = dict(value)
            # Remove default values
            value_dict = {
                field: field_value
                for field, field_value in value_dict.items()
                if not is_default_value(value, field, field_value)
            }
            return [[k, self.to_nested_list(v)] for k, v in value_dict.items()]



import logging

from rpft.parsers.creation.models import TemplateSheet


LOGGER = logging.getLogger(__name__)


def map_template_arguments(template: TemplateSheet, args, context, data_sheets) -> dict:
    """
    Map template arguments, which are positional, to the arguments from the template
    definition, and add the values of the arguments to the context with the appropriate
    variable name (from the definition).
    """
    arg_defs = template.argument_definitions

    if arg_defs and len(args) > len(arg_defs):
        # Once the row parser is cleaned up to eliminate trailing '' entries, this
        # won't be necessary
        extra_args = args[len(arg_defs) :]
        non_empty_extra_args = [ea for ea in extra_args if ea]

        if non_empty_extra_args:
            LOGGER.warning(
                "Too many template arguments provided, "
                + str(
                    {
                        "template": template.name,
                        "extra": non_empty_extra_args,
                        "definition": arg_defs,
                        "arguments": args,
                    }
                )
            )

        args = args[: len(arg_defs)]

    args_padding = [""] * (len(arg_defs) - len(args))

    for arg_def, arg in zip(arg_defs, args + args_padding):
        value = arg if arg != "" else arg_def.default_value

        if value == "":
            raise Exception(f'Required template argument "{arg_def.name}" not provided')

        value = data_sheets[value].rows if arg_def.type == "sheet" else value

        if arg_def.name in context and value != context[arg_def.name]:
            LOGGER.warn(
                "Template argument reassigned, "
                + str(
                    {
                        "template": template.name,
                        "name": arg_def.name,
                        "before": context[arg_def.name],
                        "after": value,
                    }
                )
            )

        context[arg_def.name] = value

    return context



import copy
import logging
from collections import defaultdict

from rpft.logger.logger import logging_context
from rpft.parsers.common.cellparser import CellParser
from rpft.parsers.common.rowparser import RowParser
from rpft.parsers.common.sheetparser import SheetParser
from rpft.parsers.creation import map_template_arguments
from rpft.parsers.creation.flowrowmodel import (
    Condition,
    Edge,
    FlowRowModel,
    list_of_pairs_to_dict,
)
from rpft.rapidpro.models.actions import (
    AddContactGroupAction,
    AddContactURNAction,
    Group,
    RemoveContactGroupAction,
    SendMessageAction,
    SetContactFieldAction,
    SetContactPropertyAction,
    SetRunResultAction,
    WhatsAppMessageTemplating,
)
from rpft.rapidpro.models.containers import FlowContainer, RapidProContainer
from rpft.rapidpro.models.exceptions import RapidProRouterError
from rpft.rapidpro.models.nodes import (
    BasicNode,
    CallWebhookNode,
    EnterFlowNode,
    RandomRouterNode,
    SwitchRouterNode,
    TransferAirtimeNode,
)
from rpft.rapidpro.models.routers import SwitchRouter


LOGGER = logging.getLogger(__name__)


def string_to_int_or_float(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


class NodeGroup:
    # NodeGroups may have multiple exit nodes.
    # add_exit connects ALL of them, except for those
    # which are marked as hard exits
    # (which are not tracked as loose_exits)

    def __init__(self):
        # node_groups may contain NodeGroups and RowNodeGroups
        self.node_groups = []

    def is_empty(self):
        return self.node_groups == []

    def entry_node(self):
        return self.node_groups[0].entry_node()

    def last_node_group(self):
        return self.node_groups[-1]

    def has_loose_exits(self):
        return any(group.has_loose_exits() for group in self.node_groups)

    def add_exit(self, destination_uuid, condition):
        if condition != Condition():
            raise Exception("Cannot attach conditional edges to a block.")
        if not self.has_loose_exits():
            raise Exception("Block has no loose exit to connect to.")
        for node_group in self.node_groups:
            if node_group.has_loose_exits():
                node_group.connect_loose_exits(destination_uuid)

    def connect_loose_exits(self, destination_uuid):
        for node_group in self.node_groups:
            node_group.connect_loose_exits(destination_uuid)

    def add_node_group(self, node_group):
        # Node: Connecting of exits is not done here.
        self.node_groups.append(node_group)

    def add_nodes_to_flow(self, flow_container):
        for node in self.node_groups:
            node.add_nodes_to_flow(flow_container)


class NoOpNodeGroup:
    # TODO: Support ui_pos, in case there's a node.

    class ParentEdge:
        def __init__(self, source_node_group, condition):
            self.source_node_group = source_node_group
            self.condition = condition

    def __init__(self):
        self.parent_edges = []
        self.router_node = None

    def has_loose_exits(self):
        if self.router_node:
            return any(
                exit.destination_uuid is None for exit in self.router_node.get_exits()
            )
        else:
            return any(
                edge.source_node_group.has_loose_exits() for edge in self.parent_edges
            )

    def connect_loose_exits(self, destination_uuid):
        if self.router_node:
            for exit in self.router_node.get_exits():
                if exit.destination_uuid is None:
                    exit.destination_uuid = destination_uuid
        else:
            for edge in self.parent_edges:
                edge.source_node_group.connect_loose_exits(destination_uuid)

    def entry_node(self):
        raise Exception(
            "NotImplementedError: go_to not implemented to link to no_op row."
        )

    def add_parent_edge(self, source_node_group, condition):
        self.parent_edges.append(NoOpNodeGroup.ParentEdge(source_node_group, condition))
        if self.router_node is not None:
            try:
                source_node_group.add_exit(self.router_node.uuid, condition)
            except RapidProRouterError as e:
                raise Exception(str(e))

    def add_exit(self, destination_uuid, condition):
        if not self.router_node:
            if condition == Condition():
                for edge in self.parent_edges:
                    edge.source_node_group.add_exit(destination_uuid, edge.condition)
                return
            else:
                if not condition.variable:
                    raise Exception("Condition must have a variable.")
                self.router_node = SwitchRouterNode(condition.variable)
                self.child_node_ref = None
                for edge in self.parent_edges:
                    # Update all the parent exits
                    edge.source_node_group.add_exit(
                        self.router_node.uuid, edge.condition
                    )

        # We now have a router_node
        # TODO: Code duplication here, especially with respect to default values.
        # Same functionality is in RowNodeGroup
        if not condition.value:
            self.router_node.update_default_exit(destination_uuid)
        else:
            self.router_node.add_choice(
                comparison_variable=condition.variable,
                comparison_type=condition.type or "has_any_word",
                comparison_arguments=[condition.value],
                category_name=condition.name,
                destination_uuid=destination_uuid,
                is_default=False,
            )

    def add_nodes_to_flow(self, flow_container):
        if self.router_node is not None:
            flow_container.add_node(self.router_node)


class RowNodeGroup:
    # Group of nodes representing a row in the sheet.
    # RowNodeGroups have a single exit node
    # (which may have multiple conditional exits)

    def __init__(self, node, row_type):
        """node: first node in the group"""
        self.nodes = [node]
        self.row_type = row_type

    def is_empty(self):
        # Constructor assumes a node
        return False

    def entry_node(self):
        return self.nodes[0]

    def add_nodes_to_flow(self, flow_container):
        for node in self.nodes:
            flow_container.add_node(node)

    def has_loose_exits(self):
        # A loose exit is a default exit (i.e. exit of a router-less node,
        # or exit connected to the default router case) which is not
        # connected to any further node.
        return any(exit.destination_uuid is None for exit in self.nodes[-1].get_exits())

    def connect_loose_exits(self, destination_uuid):
        for exit in self.nodes[-1].get_exits():
            if exit.destination_uuid is None:
                exit.destination_uuid = destination_uuid

    def add_exit(self, destination_uuid, condition):
        exit_node = self.nodes[-1]
        # Unconditional/default case edge
        if condition == Condition() and not isinstance(exit_node, RandomRouterNode):
            # For BasicNode, this updates the default exit.
            # For SwitchRouterNode, this updates the exit of the default category.
            # For EnterFlowNode, this should throw an error.
            try:
                exit_node.update_default_exit(destination_uuid)
            except ValueError as e:
                raise Exception(str(e))
            return

        # Completed/Expired edge from start_new_flow
        if isinstance(exit_node, EnterFlowNode):
            if condition.value.lower() in ["complete", "completed"]:
                exit_node.update_completed_exit(destination_uuid)
                self.has_loose_exit = False
            elif condition.value.lower() == "expired":
                exit_node.update_expired_exit(destination_uuid)
            else:
                LOGGER.error(
                    "Condition from start_new_flow must be 'Completed' or 'Expired'."
                )
            return

        # Completed/Expired edge from start_new_flow
        if isinstance(exit_node, CallWebhookNode) or isinstance(
            exit_node, TransferAirtimeNode
        ):
            if condition.value.lower() == "success":
                exit_node.update_success_exit(destination_uuid)
                self.has_loose_exit = False
            elif condition.value.lower() == "failure":
                exit_node.update_failure_exit(destination_uuid)
            else:
                LOGGER.error(
                    "Condition from call_webhook/transfer_airtime must be "
                    "'Success' or 'Failure'."
                )
            return

        # No Response edge from wait_for_response
        if (
            isinstance(exit_node, SwitchRouterNode)
            and condition.value.lower() == "no response"
        ):
            if exit_node.has_positive_wait():
                exit_node.update_no_response_exit(destination_uuid)
            else:
                LOGGER.warn(
                    "No Response exit only exists for wait_for_response rows with"
                    " non-zero timeout."
                )
            return

        # We have a non-trivial condition. Fill in default values if necessary
        condition_type = condition.type
        comparison_arguments = [condition.value]
        wait_timeout = None
        if self.row_type in ["split_by_group", "split_by_value"]:
            variable = exit_node.router.operand
            if self.row_type == "split_by_group":
                # Should such defaults be initialized as part of the data model?
                condition_type = "has_group"
                # TODO: Validation step that fills in group/flow uuids
                comparison_arguments = [None, condition.value]
        elif condition.variable:
            variable = condition.variable
        else:
            # TODO: Check if the source node has a save_name, and use that instead?
            wait_timeout = 0
            variable = "@input.text"

        if isinstance(exit_node, BasicNode):
            # We have a basic node, but a non-trivial condition.
            # Create a router node (new exit_node) and connect it.
            old_exit_node = exit_node
            old_destination = old_exit_node.default_exit.destination_uuid
            exit_node = SwitchRouterNode(variable, wait_timeout=wait_timeout)
            exit_node.update_default_exit(old_destination)
            old_exit_node.update_default_exit(exit_node.uuid)
            self.nodes.append(exit_node)

        # Add an outgoing edge to the router
        if isinstance(exit_node.router, SwitchRouter):
            exit_node.add_choice(
                comparison_variable=variable,
                comparison_type=condition_type or "has_any_word",
                comparison_arguments=comparison_arguments,
                category_name=condition.name,
                destination_uuid=destination_uuid,
                is_default=False,
            )
        else:  # Random router
            # TODO: If a value is provided, respect the ordering provided there
            exit_node.add_choice(
                category_name=condition.name or condition.value,
                destination_uuid=destination_uuid,
            )


class FlowParser:
    def __init__(
        self,
        rapidpro_container,
        flow_name,
        table=None,
        flow_uuid=None,
        context=None,
        sheet_parser=None,
        definition=None,
        flow_type=None,
    ):
        """
        rapidpro_container: The parent RapidProContainer to contain the flow generated
            by this parser.
        flow_name: Name to be given to the flow.
        table: tablib.Dataset: The sheet rows generating the flow;
            either table or sheet_parser must be provided
        flow_uuid: UUID to be given to the flow.
        context: Context to be used when instantiating templates in the flow.
        sheet_parser: SheetParser to be used to generate the flow;
            note that a SheetParser contains a table by iself, and if this argument is
            provided, the table argument is ignored.
        """

        self.rapidpro_container = rapidpro_container
        self.flow_name = flow_name
        self.flow_uuid = flow_uuid
        self.flow_type = flow_type or "messaging"
        self.context = context or {}
        if sheet_parser:
            self.sheet_parser = sheet_parser
        else:
            assert table is not None
            self.sheet_parser = SheetParser(
                table,
                row_parser=RowParser(FlowRowModel, CellParser()),
                context=self.context,
            )
        self.node_group_stack = [NodeGroup()]
        self.row_id_to_nodegroup = defaultdict()
        self.node_name_to_node_map = defaultdict()
        self.definition = definition

    def current_node_group(self):
        # New stuff that's created is always added to the current NodeGroup,
        # i.e. the one at the top of the stack.
        return self.node_group_stack[-1]

    def most_recent_node_group(self):
        # None is returned if no row has been processed yet.
        for stack_entry in reversed(self.node_group_stack):
            if not stack_entry.is_empty():
                return stack_entry.last_node_group()
        return None

    def append_node_group(self, new_node_group, row_id):
        self.current_node_group().add_node_group(new_node_group)
        if row_id:
            self.row_id_to_nodegroup[row_id] = new_node_group

    def parse_as_block(self):
        self._parse_block()
        if not len(self.node_group_stack) == 1:
            raise Exception("Unexpected end of flow. Did you forget end_for/end_block?")
        return self.current_node_group()

    def parse(self, add_to_container=True):
        self._parse_block()
        flow_container = self._compile_flow()
        if add_to_container:
            self.rapidpro_container.add_flow(flow_container)
        return flow_container

    def _parse_block(self, depth=0, block_type="root_block", omit_content=False):
        row, row_idx = self.sheet_parser.parse_next_row(
            omit_templating=omit_content,
            return_index=True,
        )

        while not self._is_end_of_block(block_type, row):
            if omit_content or not row.include_if:
                if row.type == "begin_for":
                    self._parse_block(depth + 1, "for", omit_content=True)
                elif row.type == "begin_block":
                    self._parse_block(depth + 1, "block", omit_content=True)
            else:
                if row.type == "begin_for":
                    if not row.mainarg_iterlist:
                        self._parse_block(depth + 1, "for", omit_content=True)

                    if len(row.loop_variable) >= 1 and row.loop_variable[0]:
                        iteration_variable = row.loop_variable[0]
                    else:
                        with logging_context(f"row {row_idx}"):
                            raise Exception("begin_for must have a loop_variable")

                    index_variable = None
                    if len(row.loop_variable) >= 2 and row.loop_variable[1]:
                        index_variable = row.loop_variable[1]

                    self.sheet_parser.create_bookmark(str(depth))
                    new_node_group = NodeGroup()
                    self.node_group_stack.append(new_node_group)

                    # Interpret the row like a no-op to get the edges
                    if not row.is_starting_row():
                        self._parse_noop_row(row, store_row_id=False)

                    for i, entry in enumerate(row.mainarg_iterlist):
                        self.sheet_parser.go_to_bookmark(str(depth))
                        self.sheet_parser.add_to_context(iteration_variable, entry)
                        if index_variable:
                            self.sheet_parser.add_to_context(index_variable, i)
                        self._parse_block(depth + 1, "for")

                    self.node_group_stack.pop()
                    self.append_node_group(new_node_group, row.row_id)
                    self.sheet_parser.remove_from_context(iteration_variable)

                    if index_variable:
                        self.sheet_parser.remove_from_context(index_variable)

                    self.sheet_parser.remove_bookmark(str(depth))
                elif row.type == "begin_block":
                    new_node_group = NodeGroup()
                    self.node_group_stack.append(new_node_group)

                    # Interpret the row like a no-op to get the edges
                    if not row.is_starting_row():
                        self._parse_noop_row(row, store_row_id=False)

                    self._parse_block(depth + 1, "block")
                    self.node_group_stack.pop()
                    self.append_node_group(new_node_group, row.row_id)
                else:
                    with logging_context(f"row {row_idx}"):
                        self._parse_row(row)

            row, row_idx = self.sheet_parser.parse_next_row(
                omit_templating=omit_content,
                return_index=True,
            )

    def _is_end_of_block(self, block_type, row):
        block_end_map = {
            "end_for": "for",
            "end_block": "block",
        }
        if row is None:
            if block_type == "root_block":
                return True
            else:
                raise Exception("Sheet has unterminated block.")
        elif row.type in block_end_map:
            if block_end_map[row.type] == block_type:
                return True
            else:
                raise Exception(
                    f'Wrong block terminator "{row.type}" found for block of type'
                    f" {block_type}."
                )
        return False

    def _parse_insert_as_block_row(self, row):
        node_group = self.get_node_group(
            row.mainarg_flow_name,
            row.data_sheet,
            row.data_row_id,
            row.template_arguments,
            self.context,
        )

        for edge in row.edges:
            self._add_row_edge(edge, node_group.entry_node().uuid)

        self.append_node_group(node_group, row.row_id)

        if row.row_id:
            self.row_id_to_nodegroup[row.row_id] = node_group

    def _get_row_action(self, row):
        if row.type == "send_message":
            templating = None
            if row.wa_template.name:
                templating = WhatsAppMessageTemplating.from_whats_app_templating_model(
                    row.wa_template
                )
            send_message_action = SendMessageAction(
                text=row.mainarg_message_text, templating=templating
            )
            for att_type, attachment in zip(
                ["image", "audio", "video"], [row.image, row.audio, row.video]
            ):
                attachment = attachment.strip()
                if attachment:
                    # TODO: Add depending on prefix.
                    send_message_action.add_attachment(f"{att_type}:{attachment}")
            for attachment in row.attachments:
                send_message_action.add_attachment(attachment)

            quick_replies = [qr for qr in row.choices if qr]
            if quick_replies:
                for qr in quick_replies:
                    send_message_action.add_quick_reply(qr)
            return send_message_action
        elif row.type == "save_value":
            return SetContactFieldAction(
                field_name=row.save_name, value=row.mainarg_value
            )
        elif row.type == "add_to_group":
            return AddContactGroupAction(
                groups=[self._get_or_create_group(row.mainarg_groups[0], row.obj_id)]
            )
        elif row.type == "add_contact_urn":
            return AddContactURNAction(
                path=row.mainarg_value,
                scheme=row.urn_scheme or "tel",
            )
        elif row.type == "remove_from_group":
            if not row.mainarg_groups:
                LOGGER.warning("Removing contact from ALL groups.")
                return RemoveContactGroupAction(groups=[], all_groups=True)
            elif row.mainarg_groups[0] == "ALL":
                return RemoveContactGroupAction(groups=[], all_groups=True)
            else:
                return RemoveContactGroupAction(
                    groups=[
                        self._get_or_create_group(row.mainarg_groups[0], row.obj_id)
                    ]
                )
        elif row.type == "save_flow_result":
            return SetRunResultAction(
                row.save_name, row.mainarg_value, category=row.result_category
            )
        elif row.type.startswith("set_contact_"):
            property = row.type.replace("set_contact_", "")

            if property not in ["channel", "language", "name", "status", "timezone"]:
                LOGGER.error(f"Unknown operation set_contact_{property}.")

            return SetContactPropertyAction(property, row.mainarg_value)
        elif row.type in [
            "wait_for_response",
            "split_by_value",
            "split_by_group",
            "split_random",
            "start_new_flow",
            "call_webhook",
            "transfer_airtime",
        ]:
            return None
        else:
            LOGGER.error(f"Row type {row.type} not implemented")

    def _get_or_create_group(self, name, uuid=None):
        # TODO: support lists of groups
        if not uuid:
            # This shouldn't be necessary, but we keep it until we
            # have tests checking for group UUIDs.
            uuid = None
        return Group(name=name, uuid=uuid)

    def _get_row_node(self, row):
        if (
            row.type in ["add_to_group", "remove_from_group", "split_by_group"]
            and row.obj_id
            and row.mainarg_groups
        ):
            self.rapidpro_container.record_group_uuid(row.mainarg_groups[0], row.obj_id)

        # TODO: Consider whether to put the functionality of getting
        # a node/action from a row should go into the Node/Action model,
        # considering that the reverse naturally is there as well.
        node_uuid = row.node_uuid or None
        if row.ui_position:
            try:
                ui_pos = [
                    int(coord) for coord in row.ui_position
                ]  # List[str] -> List[int]
            except ValueError:
                LOGGER.warn("UI position coordinates have to be integers")
                ui_pos = None

            if not ui_pos or len(ui_pos) != 2:
                LOGGER.warn("A UI position must consist of exactly two coordiates")
                ui_pos = None
        else:
            ui_pos = None

        if row.type in [
            "send_message",
            "save_value",
            "add_to_group",
            "remove_from_group",
            "save_flow_result",
        ]:
            node = BasicNode(uuid=node_uuid, ui_pos=ui_pos)
            node.update_default_exit(None)
            return node
        elif row.type in ["start_new_flow"]:
            if row.obj_id:
                self.rapidpro_container.record_flow_uuid(
                    row.mainarg_flow_name, row.obj_id
                )
            return EnterFlowNode(row.mainarg_flow_name, uuid=node_uuid, ui_pos=ui_pos)
        elif row.type in ["call_webhook"]:
            try:
                headers = list_of_pairs_to_dict(row.webhook.headers)
            except ValueError as e:
                raise Exception("webhook.headers: " + str(e))
            return CallWebhookNode(
                result_name=row.save_name,
                url=row.webhook.url,
                method=row.webhook.method,
                body=row.webhook.body,
                headers=headers,
                uuid=node_uuid,
                ui_pos=ui_pos,
            )
        elif row.type in ["transfer_airtime"]:
            try:
                airtime_amounts = list_of_pairs_to_dict(row.mainarg_dict)
            except ValueError as e:
                raise Exception("airtime_amounts: " + str(e))

            try:
                airtime_amounts = {
                    k: string_to_int_or_float(v) for k, v in airtime_amounts.items()
                }
            except ValueError:
                raise Exception("airtime_amounts: Current values must be numerical")

            return TransferAirtimeNode(
                result_name=row.save_name,
                amounts=airtime_amounts,
                uuid=node_uuid,
                ui_pos=ui_pos,
            )
        elif row.type in ["wait_for_response"]:
            if row.no_response:
                return SwitchRouterNode(
                    "@input.text",
                    result_name=row.save_name,
                    wait_timeout=int(row.no_response),
                    uuid=node_uuid,
                    ui_pos=ui_pos,
                )
            else:
                return SwitchRouterNode(
                    "@input.text",
                    result_name=row.save_name,
                    wait_timeout=0,
                    uuid=node_uuid,
                    ui_pos=ui_pos,
                )
        elif row.type in ["split_by_value"]:
            return SwitchRouterNode(
                row.mainarg_expression,
                result_name=row.save_name,
                wait_timeout=None,
                uuid=node_uuid,
                ui_pos=ui_pos,
            )
        elif row.type in ["split_by_group"]:
            return SwitchRouterNode(
                "@contact.groups",
                result_name=row.save_name,
                wait_timeout=None,
                uuid=node_uuid,
                ui_pos=ui_pos,
            )
        elif row.type in ["split_random"]:
            return RandomRouterNode(
                result_name=row.save_name, uuid=node_uuid, ui_pos=ui_pos
            )
        else:
            return BasicNode(uuid=node_uuid, ui_pos=ui_pos)

    def _get_node_name(self, row):
        return row.node_uuid or row.node_name

    def _get_node_group_from_edge(self, edge):
        if edge.from_ == "start":
            return None
        elif edge.from_:
            if edge.from_ not in self.row_id_to_nodegroup:
                raise Exception(
                    f'Edge from row_id "{edge.from_}" which does not exist.'
                )
            return self.row_id_to_nodegroup[edge.from_]
        else:
            return self.most_recent_node_group()

    def _add_row_edge(self, edge, destination_uuid):
        from_node_group = self._get_node_group_from_edge(edge)
        if from_node_group is not None:
            try:
                from_node_group.add_exit(destination_uuid, edge.condition)
            except RapidProRouterError as e:
                raise Exception(str(e))

    def _parse_goto_row(self, row):
        # If there is a single destination, connect all edges to that destination.
        # If not, ensure the number of edges and destinations match.
        destination_row_ids = (
            row.mainarg_destination_row_ids * len(row.edges)
            if len(row.mainarg_destination_row_ids) == 1
            else row.mainarg_destination_row_ids
        )

        if not len(row.edges) == len(destination_row_ids):
            raise Exception(
                "If a go_to has multiple destinations, the number of destinations has"
                " to match the number of incoming edges."
            )

        for edge, destination_row_id in zip(row.edges, destination_row_ids):
            self._add_row_edge(
                edge,
                self.row_id_to_nodegroup[destination_row_id].entry_node().uuid,
            )

    def _parse_noop_row(self, row, store_row_id=True):
        new_node = NoOpNodeGroup()
        for edge in row.edges:
            source_node_group = self._get_node_group_from_edge(edge)
            if source_node_group is not None:
                new_node.add_parent_edge(source_node_group, edge.condition)
        self.append_node_group(new_node, row.row_id if store_row_id else "")

    def _parse_row(self, row):
        if not row.include_if:
            return

        if row.type in ["hard_exit", "loose_exit"]:
            destination_uuid = "HARD_EXIT" if row.type == "hard_exit" else None
            for edge in row.edges:
                self._add_row_edge(edge, destination_uuid)
            return

        if row.type == "go_to":
            self._parse_goto_row(row)
            return

        if row.type == "no_op":
            self._parse_noop_row(row)
            return

        if row.type == "insert_as_block":
            self._parse_insert_as_block_row(row)
            return

        row_action = self._get_row_action(row)
        node_name = self._get_node_name(row)
        existing_node = self.node_name_to_node_map.get(node_name)

        if node_name and existing_node and row_action:
            # If we want to add an action to an existing node of a given name,
            # there must be exactly one unconditional edge, and the
            # from_ row_id has to match.
            if len(row.edges) == 1 and row.edges[0].condition == Condition():
                # Get the group that this row is linked to.
                predecessor_group = (
                    self.row_id_to_nodegroup.get(row.edges[0].from_)
                    if row.edges[0].from_
                    else self.most_recent_node_group()
                )

                if predecessor_group.entry_node() == existing_node:
                    existing_node.add_action(row_action)
                    if row.row_id:
                        self.row_id_to_nodegroup[row.row_id] = self.row_id_to_nodegroup[
                            row.edges[0].from_
                        ]
                    return
                else:
                    raise Exception(
                        f'To merge rows using node name "{node_name}" into a single'
                        f' node, edge must come from a node with name "{node_name}".'
                    )
            else:
                raise Exception(
                    f'To merge rows using node name "{node_name}" into a single node,'
                    " there must be exactly one unconditional incoming edge."
                )

        new_node = self._get_row_node(row)
        if row_action:
            new_node.add_action(row_action)

        for i, edge in enumerate(row.edges):
            if edge != Edge() or i == 0:
                # Omit trivial edges (i.e. from is blank so implicitly goes
                # (from the previous row, and no condition either)
                # unless it is the first edge in the list.
                self._add_row_edge(edge, new_node.uuid)

        new_node_group = RowNodeGroup(new_node, row.type)
        self.append_node_group(new_node_group, row.row_id)
        self.node_name_to_node_map[self._get_node_name(row)] = new_node

    def _compile_flow(self):
        """
        Referenced groups or other flows may have a UUID which is None.
        Add the flow to a RapidProContainer and run update_global_uuids
        to fill in these missing UUIDs in a consistent way.
        """

        flow_container = FlowContainer(
            flow_name=self.flow_name, uuid=self.flow_uuid, type=self.flow_type
        )
        if not len(self.node_group_stack) == 1:
            raise Exception("Unexpected end of flow. Did you forget end_for/end_block?")
        self.current_node_group().add_nodes_to_flow(flow_container)
        return flow_container

    def get_node_group(
        self,
        template_name,
        data_sheet,
        data_row_id,
        template_arguments,
        context=None,
    ):
        if (data_sheet and data_row_id) or (not data_sheet and not data_row_id):
            with logging_context(f"{template_name}"):
                return FlowParser._parse_flow(
                    template_name,
                    data_sheet,
                    data_row_id,
                    template_arguments,
                    RapidProContainer(),
                    context=context or {},
                    parse_as_block=True,
                    definition=self.definition,
                )
        else:
            raise Exception(
                "For insert_as_block, either both data_sheet and data_row_id or neither"
                " have to be provided."
            )

    def _parse_flow(
        sheet_name,
        data_sheet,
        data_row_id,
        template_arguments,
        rapidpro_container,
        new_name="",
        parse_as_block=False,
        context=None,
        definition=None,
        flow_type=None,
    ):
        base_name = new_name or sheet_name
        context = copy.copy(context or {})

        if data_sheet and data_row_id:
            flow_name = " - ".join([base_name, data_row_id])
            context.update(dict(definition.get_data_sheet_row(data_sheet, data_row_id)))
        else:
            if data_sheet or data_row_id:
                LOGGER.warn(
                    "For create_flow, if no data_sheet is provided, "
                    "data_row_id should be blank as well."
                )

            flow_name = base_name

        template_sheet = definition.get_template(sheet_name)
        context = map_template_arguments(
            template_sheet,
            template_arguments,
            context,
            definition.data_sheets,
        )
        flow_parser = FlowParser(
            rapidpro_container,
            flow_name,
            template_sheet.table,
            context=context,
            definition=definition,
            flow_type=flow_type,
        )

        if parse_as_block:
            return flow_parser.parse_as_block()
        else:
            return flow_parser.parse(add_to_container=False)

    @classmethod
    def parse_all(cls, definition, rapidpro_container):
        flows = {}

        for logging_prefix, row in definition.flow_definitions:
            with logging_context(f"{logging_prefix} | {row.sheet_name[0]}"):
                flow_type = row.options.get("flow_type") or "messaging"
                if row.data_sheet and not row.data_row_id:
                    data_rows = definition.get_data_sheet_rows(row.data_sheet)

                    for data_row_id in data_rows.keys():
                        with logging_context(f'with data_row_id "{data_row_id}"'):
                            flow = cls._parse_flow(
                                row.sheet_name[0],
                                row.data_sheet,
                                data_row_id,
                                row.template_arguments,
                                rapidpro_container,
                                row.new_name,
                                context=definition.global_context,
                                definition=definition,
                                flow_type=flow_type,
                            )

                            if flow.name in flows:
                                LOGGER.warning(
                                    f"Multiple definitions of flow '{flow.name}'. "
                                    "Overwriting."
                                )

                            flows[flow.name] = flow
                elif not row.data_sheet and row.data_row_id:
                    raise Exception(
                        "For create_flow, if data_row_id is provided, data_sheet must"
                        " also be provided."
                    )
                else:
                    flow = cls._parse_flow(
                        row.sheet_name[0],
                        row.data_sheet,
                        row.data_row_id,
                        row.template_arguments,
                        rapidpro_container,
                        row.new_name,
                        context=definition.global_context,
                        definition=definition,
                        flow_type=flow_type,
                    )

                    if flow.name in flows:
                        LOGGER.warning(
                            f"Multiple definitions of flow '{flow.name}'. "
                            "Overwriting."
                        )

                    flows[flow.name] = flow

        for flow in flows.values():
            rapidpro_container.add_flow(flow)



import re

from rpft.rapidpro.models.exceptions import RapidProActionError
from rpft.rapidpro.utils import generate_new_uuid


FIELD_KEY_MAX_LENGTH = 36
RESERVED_FIELD_KEYS = [
    "created_by",
    "created_on",
    "discord",
    "ext",
    "facebook",
    "fcm",
    "first_name",
    "freshchat",
    "groups",
    "has",
    "id",
    "instagram",
    "is",
    "jiochat",
    "language",
    "line",
    "mailto",
    "modified_by",
    "name",
    "rocketchat",
    "scheme",
    "tel",
    "telegram",
    "twitter",
    "twitterid",
    "uuid",
    "viber",
    "vk",
    "wechat",
    "whatsapp",
]


def mangle_string(string):
    string = re.sub(r"[. ]", "_", string)
    string = re.sub(r"[^A-Za-z0-9\_\-]+", "", string)
    return string[:15]


class Exit:
    def __init__(self, destination_uuid=None, uuid=None):
        self.uuid = uuid if uuid else generate_new_uuid()
        self.destination_uuid = destination_uuid

    def from_dict(data):
        return Exit(**data)

    def is_hard_exit(self):
        # This is a notion introduced in the context of blocks,
        # which are groups of nodes included in a flow.
        # By default, when attaching nodes to the end of a block, all
        # unconnected exits from the block are connected, however,
        # hard exits are omitted and exit the flow.
        if self.destination_uuid == "HARD_EXIT":
            return True
        return False

    def render(self):
        destination_uuid = self.destination_uuid
        if self.destination_uuid == "HARD_EXIT":
            destination_uuid = None
        return {
            "destination_uuid": destination_uuid,
            "uuid": self.uuid,
        }


class FlowReference:
    def from_dict(data):
        return FlowReference(**data)

    def __init__(self, name, uuid=None):
        self.name = name
        self.uuid = uuid

    def record_uuid(self, uuid_dict):
        uuid_dict.record_flow_uuid(self.name, self.uuid)

    def assign_uuid(self, uuid_dict):
        self.uuid = uuid_dict.get_flow_uuid(self.name)

    def render(self):
        return {"name": self.name, "uuid": self.uuid}


def generate_field_key(field_name):
    key = field_name.strip().lower().replace(" ", "_")

    if len(key) > FIELD_KEY_MAX_LENGTH:
        raise RapidProActionError(
            "Contact field key length limit exceeded",
            {"length": len(key), "limit": FIELD_KEY_MAX_LENGTH, "key": key},
        )

    if not re.match(r"^[a-z][a-z0-9_]*$", key):
        raise RapidProActionError(
            "Contact field key needs to start with a letter",
            {"key": key, "name": field_name},
        )

    return key


def generate_field_name(field_name):
    # Field names may not contain underscores
    return field_name.replace("_", " ")


class ContactFieldReference:

    def __init__(self, name, key=None, value_type=None):
        self.key = key or generate_field_key(name)
        self.name = generate_field_name(name)
        self.value_type = value_type

    @classmethod
    def from_dict(cls, data):
        return cls(**data)

    def render(self):
        data = {"key": self.key}

        if self.value_type:
            data["type"] = self.value_type

        return data


class SystemContactField(ContactFieldReference):

    def render(self):
        data = super().render()
        data["label"] = self.name

        return data


class UserContactField(ContactFieldReference):

    def __init__(self, name, key=None, value_type=None):
        super().__init__(name, key, value_type)

        if self.key in RESERVED_FIELD_KEYS:
            raise RapidProActionError(
                "Reserved name used as user contact field key",
                {"name": name, "key": self.key},
            )

    def render(self):
        data = super().render()
        data["name"] = self.name

        return data


class Group:
    def from_dict(data):
        return Group(**data)

    def __init__(
        self, name, uuid=None, query=None, status=None, system=None, count=None
    ):
        self.name = name
        self.uuid = uuid
        self.query = query
        self.status = status
        self.system = system
        self.count = count

    def record_uuid(self, uuid_dict):
        uuid_dict.record_group_uuid(self.name, self.uuid)

    def assign_uuid(self, uuid_dict):
        self.uuid = uuid_dict.get_group_uuid(self.name)

    def render(self):
        render_dict = {"name": self.name, "uuid": self.uuid}
        for attribute in ["query", "status", "system", "count"]:
            value = getattr(self, attribute)
            if value is not None:
                render_dict[attribute] = value
        return render_dict



import json
import logging
from pathlib import Path

from benedict import benedict
from tablib import Dataset

from rpft.parsers.universal import tabulate
from rpft.parsers.common.model_inference import model_from_headers
from rpft.parsers.common.sheetparser import SheetParser
from rpft.parsers.sheets import Sheet


LOGGER = logging.getLogger(__name__)


class JSONDataSource:

    def __init__(self, paths):
        self.objs = []

        for path in paths:
            with open(path, "r") as file_:
                self.objs += [(json.load(file_), Path(path).name)]

    def get(self, key, model=None):
        candidates = []

        for obj, name in self.objs:
            if key in obj:
                candidates += [(obj[key], name)]

        if not candidates:
            raise Exception("Data for key not found", {"key": key})

        active, name = candidates[-1]

        if len(candidates) > 1:
            LOGGER.debug(
                "Duplicate sheets found, "
                + str(
                    {
                        "name": key,
                        "readers": [name for _, name in candidates],
                        "active": name,
                    }
                ),
            )

        model = model or benedict

        return [model(**item) for item in active], name, key

    def get_all(self, key, model=None):
        model = model or benedict
        items = []

        for obj, name in self.objs:
            if key in obj:
                items += [([model(**item) for item in obj[key]], name, key)]

        return items

    def _get_sheet_or_die(self, sheet_name):
        data, meta, name, key = self._get(sheet_name)
        table = tabulate(data, meta)
        sheet = Sheet(None, name, Dataset(*table[1:], headers=table[0], title=key))

        return sheet

    def _get(self, key):
        candidates = []

        for obj, name in self.objs:
            if key in obj:
                candidates += [
                    (
                        obj[key],
                        obj.get("_idems", {}).get("tabulate", {}).get(key, {}),
                        name,
                    )
                ]

        if not candidates:
            raise Exception("Data for key not found", {"key": key})

        active, meta, name = candidates[-1]

        if len(candidates) > 1:
            LOGGER.debug(
                "Duplicate sheets found, "
                + str(
                    {
                        "name": key,
                        "readers": [name for *_, name in candidates],
                        "active": name,
                    }
                ),
            )

        return active, meta, name, key


class SheetDataSource:

    def __init__(self, readers):
        self.readers = readers

    def get(self, key, model=None):
        sheet = self._get_sheet_or_die(key)
        model = model or model_from_headers(key, sheet.table.headers)

        return SheetParser(sheet.table, model).parse_all(), sheet.reader.name, key

    def get_all(self, key, model=None):
        return [
            (
                SheetParser(
                    sheet.table,
                    model or model_from_headers(key, sheet.table.headers),
                ).parse_all(),
                sheet.reader.name,
                sheet.name,
            )
            for sheet in self._get_sheets_by_name(key)
        ]

    def _get_sheet_or_die(self, sheet_name):
        candidates = self._get_sheets_by_name(sheet_name)

        if not candidates:
            raise Exception("Sheet not found", {"name": sheet_name})

        active = candidates[-1]

        if len(candidates) > 1:
            readers = [c.reader.name for c in candidates]
            LOGGER.warning(
                "Duplicate sheets found, "
                + str(
                    {
                        "name": sheet_name,
                        "readers": readers,
                        "active": active.reader.name,
                    }
                ),
            )

        return active

    def _get_sheets_by_name(self, name):
        return [
            sheet
            for reader in self.readers
            if (sheet := reader.get_sheet(name)) is not None
        ]



import copy

from rpft.parsers.common.cellparser import CellParser
from rpft.parsers.common.rowdatasheet import RowDataSheet
from rpft.parsers.common.rowparser import RowParser
from rpft.parsers.creation.flowrowmodel import Edge, FlowRowModel
from rpft.rapidpro.models.actions import Group
from rpft.rapidpro.models.campaigns import Campaign
from rpft.rapidpro.models.nodes import BaseNode
from rpft.rapidpro.models.triggers import Trigger
from rpft.rapidpro.utils import generate_new_uuid


class RapidProContainer:
    def __init__(
        self,
        campaigns=None,
        fields=None,
        flows=None,
        groups=None,
        site=None,
        triggers=None,
        version="13",
    ):
        self.campaigns = campaigns or []
        self.fields = fields or []
        self.flows = flows or []
        self.groups = groups or []
        self.site = site or "https://rapidpro.idems.international"
        self.triggers = triggers or []
        self.version = version
        self.uuid_dict = UUIDDict()

    def from_dict(data):
        data_copy = copy.deepcopy(data)
        flows = data_copy.pop("flows")
        flows = [FlowContainer.from_dict(flow) for flow in flows]
        groups = data_copy.pop("groups")
        groups = [Group.from_dict(group) for group in groups]
        campaigns = data_copy.pop("campaigns")
        campaigns = [Campaign.from_dict(campaign) for campaign in campaigns]
        container = RapidProContainer(**data_copy)
        triggers = data_copy.pop("triggers")
        triggers = [Trigger.from_dict(trigger) for trigger in triggers]
        container.flows = flows
        container.groups = groups
        container.campaigns = campaigns
        container.triggers = triggers
        return container

    def add_flow(self, flow):
        self.flows.append(flow)
        self.record_flow_uuid(flow.name, flow.uuid)

    def add_flows(self, flows):
        for flow in flows:
            self.add_flow(flow)

    def add_campaign(self, campaign):
        self.campaigns.append(campaign)

    def add_trigger(self, trigger):
        self.triggers.append(trigger)

    def record_group_uuid(self, name, uuid):
        self.uuid_dict.record_group_uuid(name, uuid)

    def record_flow_uuid(self, name, uuid):
        self.uuid_dict.record_flow_uuid(name, uuid)

    def update_global_uuids(self):
        # Prefill with existings flows and groups
        for group in self.groups:
            self.uuid_dict.record_group_uuid(group.name, group.uuid)
        for flow in self.flows:
            self.uuid_dict.record_flow_uuid(flow.name, flow.uuid)

        # Update group/flow UUIDs referenced within the flows
        for flow in self.flows:
            flow.record_global_uuids(self.uuid_dict)
        for campaign in self.campaigns:
            campaign.record_global_uuids(self.uuid_dict)
        for trigger in self.triggers:
            trigger.record_global_uuids(self.uuid_dict, require_existing=True)
        self.uuid_dict.generate_missing_uuids()
        for flow in self.flows:
            flow.assign_global_uuids(self.uuid_dict)
        for campaign in self.campaigns:
            campaign.assign_global_uuids(self.uuid_dict)
        for trigger in self.triggers:
            trigger.assign_global_uuids(self.uuid_dict)

    def merge(self, container):
        """Merge another RapidPro container into this one.

        Should take the union of the flows, groups, etc, and check consistency
        of other parameters (e.g. site)
        """
        raise NotImplementedError

    def validate(self):
        self.update_global_uuids()
        self.groups = self.uuid_dict.get_group_list()
        # TODO: Update self.fields

    def render(self):
        self.validate()
        return {
            "campaigns": [campaign.render() for campaign in self.campaigns],
            "fields": self.fields,
            "flows": [flow.render() for flow in self.flows],
            "groups": [group.render() for group in self.groups],
            "site": self.site,
            "triggers": [trigger.render() for trigger in self.triggers],
            "version": self.version,
        }


class FlowContainer:
    def __init__(
        self,
        flow_name,
        type="messaging",
        language="eng",
        uuid=None,
        spec_version="13.1.0",
        revision=0,
        expire_after_minutes=10080,
        metadata=None,
        localization=None,
    ):
        # UI is not part of this as it is captured within the nodes.
        # Localization may be handled differently in the future (e.g. stored within
        # nodes or similar);
        # it is likely to be dropped from here, and only here temporarily to avoid
        # losing its data.
        self.uuid = uuid or generate_new_uuid()
        self.name = flow_name
        self.language = language
        self.type = type
        self.nodes = []
        self.spec_version = spec_version
        self.revision = revision
        self.expire_after_minutes = expire_after_minutes
        self.metadata = metadata or {}
        self.localization = localization or {}

    def from_dict(data):
        data_copy = copy.deepcopy(data)
        name = data_copy.pop("name")
        data_copy["flow_name"] = name
        nodes = data_copy.pop("nodes")
        nodes = [BaseNode.from_dict(node) for node in nodes]
        if "_ui" in data_copy:
            ui = data_copy.pop("_ui")
            if "nodes" in ui:
                for node in nodes:
                    node.add_ui_from_dict(ui["nodes"])
        flow_container = FlowContainer(**data_copy)
        flow_container.nodes = nodes
        return flow_container

    def add_node(self, node):
        self.nodes.append(node)

    def record_global_uuids(self, uuid_dict):
        for node in self.nodes:
            node.record_global_uuids(uuid_dict)

    def assign_global_uuids(self, uuid_dict):
        for node in self.nodes:
            node.assign_global_uuids(uuid_dict)

    def render(self):
        render_dict = {
            "uuid": self.uuid,
            "name": self.name,
            "language": self.language,
            "type": self.type,
            "nodes": [node.render() for node in self.nodes],
            "spec_version": self.spec_version,
            "revision": self.revision,
            "expire_after_minutes": self.expire_after_minutes,
            "metadata": self.metadata,
            "localization": self.localization,
        }
        ui_dict = {}
        for node in self.nodes:
            node_ui = node.render_ui()
            if node_ui:
                ui_dict[node.uuid] = node_ui
        if ui_dict:
            render_dict["_ui"] = {"nodes": ui_dict}
        return render_dict

    def find_node(self, uuid):
        for node in self.nodes:
            if node.uuid == uuid:
                return node
        raise ValueError(f"Destination node {uuid} does not exist within flow.")

    def _to_rows_recurse(self, node, parent_edge):
        # The version of the graph encoded in a sheet is always a DAG, if we disregard
        # all go_to edges.
        # So we effectively do the DFS version of topological sort here, with the one
        # special case that if we encounter a backward edge (and thus a cycle), we
        # convert it into a go_to edge.
        # We use temporary row_ids here (derived from node uuids) that get converted to
        # sequential ids later.
        self.visited_nodes.add(node.uuid)
        temp_row_id = f"{node.uuid}|{node.short_name()}"
        # Initiate the row model(s) for the node with one incoming edge.
        # More edges may be added later throughout the DFS
        node.initiate_row_models(temp_row_id, parent_edge)
        # Outgoing edges from the nodes (as pairs of exits with edge objects).
        # Each edge will be added as an incoming edge to the node
        # pointed to by the corresponding exit
        exits_edges = node.get_exit_edge_pairs()
        # We go backwards through the outgoing edges:
        # This way, the nodes in the right-most branch will be completed first,
        # and thus appear last in the sheet.
        for exit, edge in exits_edges[::-1]:
            if not exit.destination_uuid:
                # If the edge leads nowhere, there's no way of encoding it in the sheet
                # format.
                # In practice, this means that cases/categories from routers may be
                # dropped if they are not connected to anything.
                continue
            child_node = self.find_node(exit.destination_uuid)
            if child_node.uuid in self.completed_nodes:
                # Edge to a later node.
                # We prepend, so that in the end,
                # the edges are in the correct order again, as in this for
                # loop we go through the edges in reverse order.
                child_node.prepend_edge_to_row_models(edge)
            elif child_node.uuid in self.visited_nodes:
                # This is a backward edge to an ancestor of this node.
                child_row_id = child_node.get_row_models()[0].row_id
                child_short_id = child_row_id.split("|")[1]
                self.rows.insert(
                    0,
                    FlowRowModel(
                        row_id=f"{generate_new_uuid()}|goto.{child_short_id}",
                        type="go_to",
                        edges=[edge],
                        mainarg_destination_row_ids=[child_row_id],
                    ),
                )
            else:
                # A new node we haven't encountered yet
                self._to_rows_recurse(child_node, edge)
        self.completed_nodes.add(node.uuid)
        # Completed row get prepended to our list, in accordance
        # with the topological sort algorithm.
        # Note: These row_models may still be modified in the
        # next steps via other incoming edges.
        self.rows = node.get_row_models() + self.rows

    def to_rows(self, numbered=False):
        if not self.nodes:
            return []
        # TODO: We might want to have a dedicated model for a list of rows
        # that can also contain metadata, used for generating a sheet.
        # TODO: These attributes pollute the namespace of the class,
        # remove them or put them into a separate class.
        self.visited_nodes = set()
        self.completed_nodes = set()
        self.rows = []
        for node in self.nodes:
            node.clear_row_model()
        # Generate the list of rows (with temp row_ids)
        self._to_rows_recurse(self.nodes[0], Edge(from_="start"))
        # We now have to remap the temp row_ids to a sequence of numbers
        # Compile the remapping dict
        temp_row_id_to_row_id = {"start": "start"}
        for idx, row in enumerate(self.rows):
            if numbered:
                new_id = str(idx + 1)
            else:
                # split off the preceding UUID
                new_base_id = row.row_id.split("|")[1]
                # Append a number (if necessary) to ensure uniqueness
                new_id = new_base_id
                counter = 1
                while new_id in temp_row_id_to_row_id.values():
                    new_id = f"{new_base_id}.{counter}"
                    counter += 1
            temp_row_id_to_row_id[row.row_id] = new_id
        # Do the remapping
        for row in self.rows:
            row.row_id = temp_row_id_to_row_id[row.row_id]
            # Also remap ids when referenced in a go_to or in an edge.
            if row.type == "go_to":
                row.mainarg_destination_row_ids = [
                    temp_row_id_to_row_id[row_id]
                    for row_id in row.mainarg_destination_row_ids
                ]
            for edge in row.edges:
                edge.from_ = temp_row_id_to_row_id[edge.from_]
        return self.rows

    def to_row_data_sheet(self, strip_uuids=False, numbered=False):
        target_headers = {"edges.*.condition"}
        excluded_headers = {"obj_id", "_nodeId"} if strip_uuids else {}
        rows = self.to_rows(numbered)
        row_parser = RowParser(FlowRowModel, CellParser())
        return RowDataSheet(row_parser, rows, target_headers, excluded_headers)


class UUIDDict:
    def __init__(self, flow_dict=None, group_dict=None):
        self.flow_dict = flow_dict or {}
        self.group_dict = group_dict or {}

    def generate_missing_uuids(self):
        for k, v in self.flow_dict.items():
            if not v:
                self.flow_dict[k] = generate_new_uuid()
        for k, v in self.group_dict.items():
            if not v:
                self.group_dict[k] = generate_new_uuid()

    def record_group_uuid(self, name, uuid):
        self._record_uuid(self.group_dict, name, uuid)

    def record_flow_uuid(self, name, uuid):
        self._record_uuid(self.flow_dict, name, uuid)

    def _record_uuid(self, uuid_dict, name, uuid):
        recorded_uuid = uuid_dict.get(name)
        if recorded_uuid:
            if uuid and uuid != recorded_uuid:
                raise ValueError(
                    f"Group/Flow {name} has multiple uuids: {uuid} and {recorded_uuid}"
                )
        else:
            uuid_dict[name] = uuid

    def get_group_uuid(self, name):
        return self.group_dict[name]

    def get_flow_uuid(self, name):
        return self.flow_dict[name]

    def contains_flow(self, name):
        return name in self.flow_dict

    def get_group_list(self):
        return [Group(name, uuid) for name, uuid in self.group_dict.items()]



import unittest

from rpft.rapidpro.models.exceptions import RapidProRouterError
from rpft.rapidpro.models.routers import SwitchRouter, RandomRouter


class TestRouters(unittest.TestCase):
    def setUp(self) -> None:
        self.switch_router = SwitchRouter(
            operand="@input.text", result_name=None, wait_timeout=600
        )
        self.switch_router.add_choice(
            "@input.text",
            "has_any_word",
            ["word"],
            "Add",
            "test_destination_1",
            is_default=False,
        )
        self.switch_router.add_choice(
            "@input.text",
            "has_any_word",
            ["word"],
            "Other",
            "test_destination_2",
            is_default=True,
        )
        self.switch_router.update_no_response_category("no_response_destination_uuid")

        self.random_router = RandomRouter()
        self.random_router.add_choice("random_1", "test_destination_1")
        self.random_router.add_choice("random_2", "test_destination_2")
        self.random_router.add_choice("random_3", "test_destination_3")

    def test_switch_router_render(self):
        render_output = self.switch_router.render()

        for key in ["type", "categories", "cases", "operand", "default_category_uuid"]:
            self.assertIn(key, list(render_output.keys()))

        self.assertEqual(render_output["type"], "switch")
        self.assertEqual(render_output["operand"], "@input.text")
        self.assertEqual(render_output["cases"][0]["type"], "has_any_word")
        self.assertEqual(render_output["cases"][0]["arguments"], ["word"])

        self.assertEqual(len(render_output["categories"]), 3)
        self.assertEqual(render_output["categories"][0]["name"], "Add")

        self.assertIn("Add", [c["name"] for c in render_output["categories"]])
        self.assertIn("Other", [c["name"] for c in render_output["categories"]])
        self.assertIn("No Response", [c["name"] for c in render_output["categories"]])
        self.assertIn("wait", render_output)
        self.assertEqual(render_output["wait"]["timeout"]["seconds"], 600)
        no_response_category = [
            c for c in render_output["categories"] if c["name"] == "No Response"
        ][0]
        self.assertEqual(
            render_output["wait"]["timeout"]["category_uuid"],
            no_response_category["uuid"],
        )

        other_category_arr = [
            c for c in render_output["categories"] if c["name"] == "Other"
        ]

        self.assertEqual(
            render_output["default_category_uuid"], other_category_arr[0]["uuid"]
        )

    def test_random_router_render(self):
        render_output = self.random_router.render()

        self.assertEqual(render_output["type"], "random")
        for key in ["type", "categories"]:
            self.assertIn(key, list(render_output.keys()))

        category_names = [c["name"] for c in render_output["categories"]]

        for name in ["random_1", "random_2", "random_3"]:
            self.assertIn(name, category_names)


class TestDuplicateChoices(unittest.TestCase):
    def test_switch_router_default(self):
        switch_router = SwitchRouter(operand="@fields.field")
        switch_router.add_choice(
            "@fields.field",
            "has_any_word",
            ["other"],
            "Other",
            "test_destination_1",
            is_default=True,
        )
        switch_router.add_choice(
            "@fields.field",
            "has_any_word",
            ["word"],
            "Word",
            "test_destination_2",
            is_default=False,
        )
        # Duplicate case with different destination
        switch_router.add_choice(
            "@fields.field",
            "has_any_word",
            ["other"],
            "Other",
            "test_destination_3",
            is_default=True,
        )

        render_output = switch_router.render()
        self.assertEqual(len(render_output["cases"]), 2)
        self.assertEqual(len(render_output["categories"]), 2)
        self.assertEqual(render_output["categories"][0]["name"], "Word")
        self.assertEqual(render_output["categories"][1]["name"], "Other")
        self.assertEqual(
            switch_router.default_category.exit.destination_uuid, "test_destination_3"
        )

    def test_switch_router_blank(self):
        switch_router = SwitchRouter(
            operand="@input.text", result_name=None, wait_timeout=600
        )
        switch_router.add_choice(
            "@input.text",
            "has_any_word",
            ["word"],
            "",
            "test_destination_1",
            is_default=False,
        )
        switch_router.update_default_category("test_destination_2")
        # Duplicate case with different destination
        switch_router.add_choice(
            "@input.text",
            "has_any_word",
            ["word"],
            "",
            "test_destination_3",
            is_default=False,
        )
        switch_router.update_no_response_category("no_response_destination_uuid")

        render_output = switch_router.render()
        self.assertEqual(len(render_output["cases"]), 1)
        self.assertEqual(len(render_output["categories"]), 3)
        self.assertEqual(render_output["categories"][0]["name"], "Word")
        cats = [
            category for category in switch_router.categories if category.name == "Word"
        ]
        # There should be exactly one Word category, internally
        self.assertEqual(len(cats), 1)
        self.assertEqual(cats[0].exit.destination_uuid, "test_destination_3")


class TestNoArgsTests(unittest.TestCase):
    def test_no_args_tests(self):
        switch_router = SwitchRouter(operand="@fields.field")
        switch_router.add_choice(
            "@fields.field",
            "has_text",
            ["junk"],
            "Has Text",
            "test_destination_1",
            is_default=False,
        )

        render_output = switch_router.render()
        self.assertEqual(len(render_output["cases"]), 1)
        self.assertEqual(len(render_output["categories"]), 2)
        self.assertEqual(render_output["cases"][0]["arguments"], [])
        self.assertEqual(render_output["categories"][0]["name"], "Has Text")
        self.assertEqual(render_output["categories"][1]["name"], "Other")
        self.assertEqual(switch_router.default_category.exit.destination_uuid, None)
        self.assertEqual(
            switch_router.categories[0].exit.destination_uuid, "test_destination_1"
        )

    def test_invalid_test(self):
        switch_router = SwitchRouter(operand="@fields.field")
        with self.assertRaises(RapidProRouterError):
            switch_router.add_choice(
                "@fields.field",
                "has_junk",
                ["junk"],
                "Has Just",
                "test_destination_1",
                is_default=False,
            )



from enum import Enum

from pydantic import model_validator

from rpft.parsers.common.rowparser import ParserModel
from rpft.parsers.creation.models import SurveyConfig


class ContentIndexType(Enum):
    SURVEY = "survey"
    SURVEY_QUESTION = "survey_question"
    CONTENT_INDEX = "content_index"
    FLOW = "create_flow"
    TEMPLATE = "template_definition"
    CAMPAIGN = "create_campaign"
    TRIGGERS = "create_triggers"
    DATA_SHEET = "data_sheet"
    IGNORE = "ignore_row"
    GLOBALS = "globals"


class TemplateArgument(ParserModel):
    name: str
    type: str = ""
    default_value: str = ""


class Operation(ParserModel):
    type: str = ""
    expression: str = ""
    order: str = ""


class ContentIndexRowModel(ParserModel):
    type: str = ""
    new_name: str = ""
    sheet_name: list[str] = []
    data_sheet: str = ""
    data_row_id: str = ""
    template_argument_definitions: list[TemplateArgument] = []  # internal name
    template_arguments: list = []
    options: dict = {}
    survey_config: SurveyConfig = SurveyConfig()
    operation: Operation = Operation()
    data_model: str = ""
    group: str = ""
    status: str = ""
    tags: list[str] = []

    def field_name_to_header_name(field):
        if field == "template_argument_definitions":
            return "template_arguments"
        if field == "survey_config":
            return "config"

    def header_name_to_field_name_with_context(header, row):
        if row["type"] == ContentIndexType.TEMPLATE.value and header == "template_arguments":
            return "template_argument_definitions"
        if row["type"] == ContentIndexType.SURVEY.value and header == "config":
            return "survey_config"
        else:
            return header

    @model_validator(mode="before")
    def from_list(cls, data):
        try:
            if data["type"] == "template_definition":
                data["template_argument_definitions"] = data.pop("template_arguments")
        except Exception:
            pass

        return data



from pydantic import field_validator

from rpft.parsers.common.rowparser import ParserModel


class CampaignEventRowModel(ParserModel):
    uuid: str = ""
    offset: int
    unit: str
    event_type: str
    delivery_hour: str = ""
    message: str = ""
    relative_to: str
    start_mode: str
    flow: str = ""
    base_language: str = ""

    @field_validator("unit")
    @classmethod
    def validate_unit(cls, v):
        if v not in ["M", "H", "D", "W"]:
            raise ValueError("unit must be M (minute), H (hour), D (day) or W (week)")
        return v

    @field_validator("start_mode")
    @classmethod
    def validate_start_mode(cls, v):
        if v not in ["I", "S", "P"]:
            raise ValueError(
                "start_mode must be I (interrupt current flow), S (skip event if in"
                " flow) or P (send message and don't affect flow)"
            )
        return v

    @field_validator("event_type")
    @classmethod
    def validate_event_type(cls, v):
        if v not in ["M", "F"]:
            raise ValueError("event_type must be F (flow) or M (message)")
        return v



import unittest
from typing import List

from rpft.parsers.common.rowparser import ParserModel, RowParser
from tests.mocks import MockCellParser


class SubModel(ParserModel):
    str_field: str = ""
    list_field: list = []


class MyModel(ParserModel):
    bool_field: int = False
    int_field: int = 0
    str_field: str = ""
    list_field: list = []
    submodel_field: SubModel = SubModel()


class BoolModel(ParserModel):
    bool_field: bool = True


class TestRowParserBoolean(unittest.TestCase):
    def setUp(self):
        self.parser = RowParser(BoolModel, MockCellParser())
        self.falseModel = BoolModel(**{"bool_field": False})
        self.trueModel = BoolModel(**{"bool_field": True})

    def test_convert_false(self):
        inputs = [
            {"bool_field": "False"},
            {"bool_field": "  False  "},
            {"bool_field": "FALSE"},
            {"bool_field": "false"},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.falseModel)

    def test_convert_true(self):
        inputs = [
            {"bool_field": "True"},
            {"bool_field": "  True  "},
            {"bool_field": "TRUE"},
            {"bool_field": "true"},
            {"bool_field": "something"},
            {"bool_field": "1"},
            {"bool_field": "0"},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.trueModel)

    def test_convert_default(self):
        inp = {}
        out = self.parser.parse_row(inp)
        self.assertEqual(out, self.trueModel)


class IntModel(ParserModel):
    int_field: int = 0


class TestRowParserInt(unittest.TestCase):
    def setUp(self):
        self.parser = RowParser(IntModel, MockCellParser())
        self.twelveModel = IntModel(**{"int_field": 12})
        self.zeroModel = IntModel(**{"int_field": 0})

    def test_convert_int(self):
        inputs = [
            {"int_field": "12"},
            {"int_field": "  12  "},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.twelveModel)
        inp = {"int_field": "twelve"}
        with self.assertRaises(ValueError):
            out = self.parser.parse_row(inp)

    def test_convert_default(self):
        inp = {}
        out = self.parser.parse_row(inp)
        self.assertEqual(out, self.zeroModel)


class ListStrModel(ParserModel):
    list_field: List[str] = []


class TestRowParserListStr(unittest.TestCase):
    def setUp(self):
        self.parser = RowParser(ListStrModel, MockCellParser())
        self.emptyModel = ListStrModel(**{"list_field": []})
        self.oneModel = ListStrModel(**{"list_field": ["1"]})
        self.onetwoModel = ListStrModel(**{"list_field": ["1", "2"]})

    def test_convert_empty(self):
        inputs = [
            {},
            # {"list_field": ""},
            {"list_field": []},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.emptyModel)

    def test_convert_single_element(self):
        inputs = [
            {"list_field": ["1"]},
            # {"list_field": ["1", ""]},
            {"list_field": "1"},
            {"list_field.1": "1"},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.oneModel)

    def test_convert_two_element(self):
        inputs = [
            {"list_field": ["1", "2"]},
            {"list_field.1": "1", "list_field.2": "2"},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.onetwoModel)


class ListIntModel(ParserModel):
    list_field: List[int] = []


class TestRowParserListInt(TestRowParserListStr):
    def setUp(self):
        self.parser = RowParser(ListIntModel, MockCellParser())
        self.emptyModel = ListIntModel(**{"list_field": []})
        self.oneModel = ListIntModel(**{"list_field": [1]})
        self.onetwoModel = ListIntModel(**{"list_field": [1, 2]})


class ListModel(ParserModel):
    list_field: list = []


class TestRowParserList(TestRowParserListStr):
    def setUp(self):
        self.parser = RowParser(ListModel, MockCellParser())
        self.emptyModel = ListModel(**{"list_field": []})
        self.oneModel = ListModel(**{"list_field": ["1"]})
        self.onetwoModel = ListModel(**{"list_field": ["1", "2"]})


class DictModel(ParserModel):
    dict_field: dict = {}


class TestRowParserDict(unittest.TestCase):
    def setUp(self):
        self.parser = RowParser(DictModel, MockCellParser())

    def test_convert_empty(self):
        self.emptyModel = DictModel(**{"dict_field": {}})
        inputs = [
            {},
            {"dict_field": ""},
            {"dict_field": []},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.emptyModel)

    def test_convert_single_element(self):
        self.oneModel = DictModel(**{"dict_field": {"K": "V"}})
        inputs = [
            {"dict_field": ["K", "V"]},
            {"dict_field": [["K", "V"]]},
            {"dict_field.K": "V"},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.oneModel)

    def test_convert_two_element(self):
        self.onetwoModel = DictModel(**{"dict_field": {"K1": "V1", "K2": "V2"}})
        inputs = [
            {"dict_field": [["K1", "V1"], ["K2", "V2"]]},
            {"dict_field.K1": "V1", "dict_field.K2": "V2"},
        ]
        for inp in inputs:
            out = self.parser.parse_row(inp)
            self.assertEqual(out, self.onetwoModel)



import json
import re
from abc import ABC
from collections.abc import Mapping
from pathlib import Path

import tablib
from googleapiclient.discovery import build

from rpft.google import get_credentials


class SheetReaderError(Exception):
    pass


class Sheet:
    def __init__(self, reader, name, table):
        self.reader = reader
        self.name = name
        self.table = table

    def __repr__(self):
        return f"Sheet(name: '{self.name}')"


class AbstractSheetReader(ABC):
    @property
    def sheets(self) -> Mapping[str, Sheet]:
        return self._sheets

    def get_sheet(self, name) -> Sheet:
        return self.sheets.get(name)

    @classmethod
    def can_process(cls, path):
        return False

    def __repr__(self):
        return f"{type(self).__name__}(name: '{self.name}')"


class CSVSheetReader(AbstractSheetReader):
    def __init__(self, path):
        self.name = path
        self._sheets = {
            f.stem: Sheet(reader=self, name=f.stem, table=load_csv(f))
            for f in Path(path).glob("*.csv")
        }

    @classmethod
    def can_process(cls, location):
        return Path(location).is_dir()


class JSONSheetReader(AbstractSheetReader):
    def __init__(self, filename):
        self.name = filename
        data = load_json(filename)
        self._sheets = {}
        for name, content in data["sheets"].items():
            table = tablib.Dataset()
            table.dict = content
            self._sheets[name] = Sheet(reader=self, name=name, table=table)

    @classmethod
    def can_process(cls, location):
        if Path(location).suffix.lower() == ".json":
            return "sheets" in load_json(location)

        return False


class XLSXSheetReader(AbstractSheetReader):
    def __init__(self, filename):
        self.name = filename
        with open(filename, "rb") as table_data:
            data = tablib.Databook().load(table_data.read(), "xlsx")
        self._sheets = {}
        for sheet in data.sheets():
            self.sheets[sheet.title] = Sheet(
                reader=self,
                name=sheet.title,
                table=sanitize(sheet),
            )

    @classmethod
    def can_process(cls, location):
        return Path(location).suffix.lower() == ".xlsx"


class GoogleSheetReader(AbstractSheetReader):

    def __init__(self, spreadsheet_id):
        """
        Args:
            spreadsheet_id: You can extract it from the spreadsheed URL, like this
            https://docs.google.com/spreadsheets/d/[spreadsheet_id]/edit
        """

        self.name = spreadsheet_id

        service = build("sheets", "v4", credentials=get_credentials())
        sheet_metadata = (
            service.spreadsheets().get(spreadsheetId=spreadsheet_id).execute()
        )
        sheets = sheet_metadata.get("sheets", "")
        titles = []
        for sheet in sheets:
            title = sheet.get("properties", {}).get("title", "Sheet1")
            titles.append(title)

        result = (
            service.spreadsheets()
            .values()
            .batchGet(spreadsheetId=spreadsheet_id, ranges=titles)
            .execute()
        )

        self._sheets = {}
        for sheet in result.get("valueRanges", []):
            name = sheet.get("range", "").split("!")[0]
            if name.startswith("'") and name.endswith("'"):
                name = name[1:-1]
            content = sheet.get("values", [])
            if name in self._sheets:
                raise ValueError(f"Warning: Duplicate sheet name: {name}")
            else:
                self._sheets[name] = Sheet(
                    reader=self,
                    name=name,
                    table=self._table_from_content(content),
                )

    def _table_from_content(self, content):
        table = tablib.Dataset()
        table.headers = content[0]

        for row in content[1:]:
            table.append(self._prepare_row(row, len(table.headers)))

        return table

    def _prepare_row(self, row, max_cols):
        return pad(
            [cell.replace("\r\n", "\n") for cell in row],
            max_cols,
        )

    @classmethod
    def can_process(cls, location):
        return bool(re.fullmatch(r"[a-z0-9_-]{44}", location, re.IGNORECASE))


class DatasetSheetReader(AbstractSheetReader):
    def __init__(self, datasets, name):
        self._sheets = {d.title: Sheet(self, d.title, d) for d in datasets}
        self.name = "[datasets]"


class ODSSheetReader(AbstractSheetReader):
    def __init__(self, path):
        book = tablib.Databook()

        with open(path, "rb") as f:
            book.load(f, format="ods")

        self._sheets = {
            sheet.title: Sheet(self, sheet.title, sanitize(sheet))
            for sheet in book.sheets()
        }
        self.name = str(path)

    @classmethod
    def can_process(cls, location):
        return Path(location).suffix.lower() == ".ods"


def sanitize(sheet):
    data = tablib.Dataset()
    data.headers = sheet.headers
    # remove trailing Nones
    while data.headers and data.headers[-1] is None:
        data.headers.pop()
    for row in sheet:
        vals = tuple(str(e) if e is not None else "" for e in row)
        new_row = vals[: len(data.headers)]
        if any(new_row):
            # omit empty rows
            data.append(new_row)
    return data


def load_csv(path):
    with open(path, mode="r", encoding="utf-8") as csv:
        return tablib.import_set(csv, format="csv")


def load_json(path):
    with open(path, mode="r", encoding="utf-8") as fjson:
        data = json.load(fjson)
    return data


def pad(row, n):
    return row + ([""] * (n - len(row)))



import copy
import logging
import re

from rpft.logger.logger import logging_context
from rpft.parsers.common.rowparser import ParserModel
from rpft.parsers.creation import map_template_arguments
from rpft.parsers.creation.models import ChatbotDefinition
from rpft.parsers.creation.flowparser import FlowParser
from rpft.rapidpro.models.containers import RapidProContainer


LOGGER = logging.getLogger(__name__)


def name_to_id(name):
    return re.sub("[^a-z0-9]", "", name.lower())


def apply_to_all_str(obj, func, inplace=False):
    """
    Apply the given function to all string fields within the given nested model.

    `func` should accept a string as argument and return a string.
    """
    if isinstance(obj, str):
        return func(obj)
    elif isinstance(obj, dict):
        return {k: apply_to_all_str(v, func) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [apply_to_all_str(v, func) for v in obj]
    elif isinstance(obj, ParserModel):
        obj2 = obj if inplace else copy.deepcopy(obj)
        for k, v in obj.__dict__.items():
            setattr(obj2, k, apply_to_all_str(v, func))
        return obj2
    else:
        return obj


class SurveyQuestion:
    def __init__(
        self,
        survey_name,
        data_row,
        template_arguments=None,
        logging_prefix=None,
    ):
        self.data_row = copy.deepcopy(data_row)
        self.logging_prefix = logging_prefix
        self.template_arguments = template_arguments or []
        self.survey_name = survey_name
        self.survey_id = name_to_id(survey_name)
        self.question_id = name_to_id(self.data_row.ID)

    @property
    def ID(self):
        return self.data_row.ID

    def initialize_survey_variables(self, survey_id):
        """Initialize empty variable names with defaults."""
        if not self.data_row.variable:
            self.data_row.variable = f"sq_{survey_id}_{self.question_id}"
        if not self.data_row.completion_variable:
            self.data_row.completion_variable = f"{self.data_row.variable}_complete"

    def get_rapidpro_variables(self):
        """Get list of variables that are created in this survey."""
        return [
            self.data_row.variable,
            self.data_row.completion_variable,
        ] + [assg.variable for assg in self.data_row.postprocessing.assignments]

    def set_default_expiration_message(self, message):
        self.data_row.expiration.message = self.data_row.expiration.message or message

    def apply_prefix_renaming(self, prefix):
        """
        Apply a prefix to the names of all question variables.
        """
        self.data_row.variable = prefix + self.data_row.variable
        self.data_row.completion_variable = prefix + self.data_row.completion_variable
        for assg in self.data_row.postprocessing.assignments:
            assg.variable = prefix + assg.variable

    def apply_prefix_substitutions(self, variables, prefix):
        """
        Apply a prefix to the given question variables.

        Applies wherever any of these variables appear in the survey (condition, message
        text or elsewhere).
        """

        def replace_vars(s):
            for var in variables:
                s = re.sub(
                    f"fields.{var}([^a-zA-Z0-9_]|$)",
                    f"fields.{prefix}{var}\\1",
                    s,
                )
            return s

        apply_to_all_str(self.data_row, replace_vars, inplace=True)

    def apply_shorthand_substitutions(self, survey_id):
        """
        Replace placeholders, like '@answer', with actual values.
        """

        def replace_vars(s):
            s = s.replace("@answerid", f"{self.data_row.variable}")
            s = s.replace("@answer", f"@fields.{self.data_row.variable}")
            s = s.replace("@prefixid", f"sq_{survey_id}")
            s = s.replace("@prefix", f"@fields.sq_{survey_id}")
            s = s.replace("@surveyid", f"{survey_id}")
            s = s.replace("@questionid", f"{self.question_id}")
            return s

        apply_to_all_str(self.data_row, replace_vars, inplace=True)

    def populate_survey_variables(self, survey_id=None):
        """
        Initialize empty variable names with defaults and insert these in place of
        shorthand placeholders (e.g. @answer).
        """

        survey_id = survey_id or self.survey_id
        self.initialize_survey_variables(survey_id)
        self.apply_shorthand_substitutions(survey_id)


class Survey:
    def __init__(
        self,
        name,
        question_data_sheet,
        survey_config,
        template_arguments=None,
        logging_prefix=None,
    ):
        self.name = name
        self.survey_id = name_to_id(name)
        self.survey_config = survey_config
        self.logging_prefix = logging_prefix
        self.template_arguments = template_arguments or []
        self.questions = [
            SurveyQuestion(name, row, template_arguments, logging_prefix)
            for row in question_data_sheet.rows.values()
        ]

    def preprocess_data_rows(self):
        for question in self.questions:
            question.populate_survey_variables(self.survey_id)

        variables = self.get_rapidpro_variables()
        prefix = self.survey_config.variable_prefix

        for question in self.questions:
            question.set_default_expiration_message(
                self.survey_config.expiration_message
            )
            if prefix:
                question.apply_prefix_renaming(prefix)
            question.apply_prefix_substitutions(variables, prefix)

    def get_rapidpro_variables(self):
        """Get list of variables that are created in this survey."""
        variables = []
        for question in self.questions:
            variables += question.get_rapidpro_variables()
        return list(set(variables))

    def get_question_by_id(self, ID):
        for question in self.questions:
            if question.ID == ID:
                return question
        return None


class SurveyParser:
    QUESTION_TEMPLATE_NAME = "template_survey_question_wrapper"
    SURVEY_TEMPLATE_NAME = "template_survey_wrapper"

    def __init__(self, definition: ChatbotDefinition):
        self.definition = definition
        self.survey_template = definition.get_template(
            SurveyParser.SURVEY_TEMPLATE_NAME
        )
        self.question_template = definition.get_template(
            SurveyParser.QUESTION_TEMPLATE_NAME
        )

    @classmethod
    def parse_all(cls, definition, container: RapidProContainer):
        for survey_question in definition.survey_questions:
            SurveyParser(definition).parse_question(survey_question, container)
        for survey in definition.surveys.values():
            SurveyParser(definition).parse_survey(survey, container)

        return container

    def parse_survey(self, survey: Survey, container: RapidProContainer):
        with logging_context(f"{survey.logging_prefix} | survey {survey.name}"):
            survey.preprocess_data_rows()
            self.parse_survey_wrapper(survey, container)

            for question in survey.questions:
                with logging_context(
                    f"{survey.logging_prefix}"
                    f" | survey {survey.name}"
                    f" | question {question.ID}"
                ):
                    self.parse_question(question, container)

        return container

    def parse_question(self, question, container: RapidProContainer):
        question.populate_survey_variables()
        context = map_template_arguments(
            self.question_template,
            question.template_arguments,
            dict(question.data_row),
            self.definition.data_sheets,
        )
        flow_parser = FlowParser(
            container,
            f"survey - {question.survey_name} - question - {question.ID}",
            self.question_template.table,
            context=context,
            definition=self.definition,
        )

        flow_parser.parse()

        return container

    def parse_survey_wrapper(
        self,
        survey,
        container: RapidProContainer,
    ):
        question_data = [question.data_row for question in survey.questions]
        context = {
            "questions": question_data,
            "survey_name": survey.name,
            "survey_id": survey.survey_id,
        }
        context = map_template_arguments(
            self.survey_template,
            survey.template_arguments,
            context,
            self.definition.data_sheets,
        )
        flow_parser = FlowParser(
            container,
            f"survey - {survey.name}",
            self.survey_template.table,
            context=context,
            definition=self.definition,
        )
        flow_parser.parse()



from unittest import TestCase

from benedict import benedict
from rpft.parsers.common.rowparser import ParserModel
from rpft.sources import JSONDataSource


class TestModel(ParserModel):
    k1: str


class TestJSONDataSource(TestCase):

    def setUp(self):
        self.source = JSONDataSource([])

    def test_get_retrieves_by_key(self):
        data = {"o1": [{"k1": "v1"}]}
        self.source.objs = [(data, "data")]
        item = self.source.get("o1")

        self.assertEqual(item, ([{"k1": "v1"}], "data", "o1"))
        self.assertEqual(type(item[0][0]), benedict)

    def test_get_retrieves_by_key_and_model(self):
        data = {"o1": [{"k1": "v1"}]}
        self.source.objs = [(data, "data")]
        item = self.source.get("o1", TestModel)

        self.assertEqual(item, ([TestModel(k1="v1")], "data", "o1"))
        self.assertEqual(type(item[0][0]), TestModel)

    def test_get_retrieves_last_duplicated_item(self):
        item1 = {"o1": [{"k1": "v1"}]}
        item2 = {"o1": [{"k1": "v2"}]}
        self.source.objs = [(item1, "item1"), (item2, "item2")]
        fetched = self.source.get("o1")

        self.assertEqual(fetched, (item2["o1"], "item2", "o1"))

    def test_raise_error_when_item_missing(self):
        self.assertRaises(Exception, self.source.get, "o1")

    def test_get_all_concatenates_items_with_same_key(self):
        item1 = {"o1": [{"k1": "v1"}]}
        item2 = {"o1": [{"k1": "v2"}]}
        self.source.objs = [(item1, "item1"), (item2, "item2")]
        fetched = self.source.get_all("o1")

        self.assertEqual(
            fetched,
            [
                ([{"k1": "v1"}], "item1", "o1"),
                ([{"k1": "v2"}], "item2", "o1"),
            ],
        )


class TestLegacySheetBasedRetrieval(TestCase):

    def test_get_sheet_by_name(self):
        source = JSONDataSource([])
        source.objs = [({"o1": [{"k1": "v1"}]}, "data")]
        sheet = source._get_sheet_or_die("o1")

        self.assertEqual(sheet.name, "data")
        self.assertEqual(sheet.table.headers, ["k1"])
        self.assertEqual(sheet.table[0], ("v1",))



from rpft.parsers.creation.datarowmodel import DataRowModel
from typing import List


class ListRowModel(DataRowModel):
    # Because this defines the content of a datasheet,
    # it inherits DataRowModel, which gives it the ID column.
    messages: List[str]


class LookupRowModel(DataRowModel):
    # Because this defines the content of a datasheet,
    # it inherits DataRowModel, which gives it the ID column.
    happy: str
    sad: str
    neutral: str



import unittest
from collections import OrderedDict
from typing import List, Optional

from rpft.parsers.common.cellparser import CellParser
from rpft.parsers.common.rowparser import ParserModel, RowParser, RowParserError
from tests.mocks import MockCellParser


class ModelWithStuff(ParserModel):
    list_str: List[str] = []
    int_field: int = 0
    str_field: str = ""


class MainModel(ParserModel):
    str_field: str = ""
    model_optional: Optional[ModelWithStuff] = None
    model_default: ModelWithStuff = ModelWithStuff()
    model_list: List[ModelWithStuff] = []


input1 = MainModel()
output1_exp = {}

submodel_2_dict = {
    "list_str": ["a", "b", "c"],
    "int_field": 5,
    "str_field": "string",
}

input2 = MainModel(
    **{
        "str_field": "main model string",
        "model_default": submodel_2_dict,
    }
)

output2_exp = OrderedDict(
    [
        ("str_field", "main model string"),
        ("model_default.list_str.1", "a"),
        ("model_default.list_str.2", "b"),
        ("model_default.list_str.3", "c"),
        ("model_default.int_field", 5),
        ("model_default.str_field", "string"),
    ]
)

input3 = MainModel(
    **{
        "str_field": "main model string",
        "model_default": {"list_str": [], "int_field": 15, "str_field": ""},
    }
)

output3_exp = OrderedDict(
    [
        ("str_field", "main model string"),
        ("model_default.int_field", 15),
    ]
)

input4 = MainModel(
    **{
        "str_field": "main model string",
        "model_default": {
            "list_str": ["a", "b"],
            "int_field": 5,
            "str_field": "default",
        },
        "model_optional": {
            "list_str": ["c", "d"],
            "int_field": 10,
            "str_field": "optional",
        },
        "model_list": [
            {
                "list_str": ["A", "B", "C"],
                "str_field": "string from first model in list",
            },
            {
                "int_field": 15,
            },
        ],
    }
)

output4_exp = OrderedDict(
    [
        ("str_field", "main model string"),
        ("model_optional.list_str.1", "c"),
        ("model_optional.list_str.2", "d"),
        ("model_optional.int_field", 10),
        ("model_optional.str_field", "optional"),
        ("model_default.list_str.1", "a"),
        ("model_default.list_str.2", "b"),
        ("model_default.int_field", 5),
        ("model_default.str_field", "default"),
        ("model_list.1.list_str.1", "A"),
        ("model_list.1.list_str.2", "B"),
        ("model_list.1.list_str.3", "C"),
        ("model_list.1.str_field", "string from first model in list"),
        ("model_list.2.int_field", 15),
    ]
)


# ------ Data for remapping test cases ------
class ChildModel(ParserModel):
    some_field: str = ""

    def field_name_to_header_name(field):
        field_map = {
            "some_field": "my_field",
        }
        return field_map.get(field, field)


class ModelWithBasicRemap(ParserModel):
    str_field: str = ""
    new_str_field: str = ""
    list_field: List[str] = []
    child_field: Optional[ChildModel]

    def field_name_to_header_name(field):
        field_map = {
            "new_str_field": "str_field_2",
            "list_field": "cool_list",
        }
        return field_map.get(field, field)


input_remap = ModelWithBasicRemap(
    **{
        "str_field": "main model string",
        "new_str_field": "new string",
        "list_field": ["cool", "list"],
        "child_field": {"some_field": "some value"},
    }
)


output_remap_exp = OrderedDict(
    [
        ("str_field", "main model string"),
        ("str_field_2", "new string"),
        ("cool_list", ["cool", "list"]),
        ("child_field.my_field", "some value"),
    ]
)


class ModelWithClashingRemap(ParserModel):
    str_field: str = ""
    list_field: List[str] = []

    def field_name_to_header_name(field):
        field_map = {
            "str_field": "new_field",
            "list_field": "new_field",
        }
        return field_map.get(field, field)


input_clashing_remap1 = ModelWithClashingRemap(
    **{
        "str_field": "string",
        "list_field": [],
    }
)


input_clashing_remap2 = ModelWithClashingRemap(
    **{
        "str_field": "",
        "list_field": ["cool", "list"],
    }
)


input_clashing_remap3 = ModelWithClashingRemap(
    **{
        "str_field": "string",
        "list_field": ["cool", "list"],
    }
)


output_clashing_remap_exp1 = {"new_field": "string"}
output_clashing_remap_exp2 = {"new_field": ["cool", "list"]}


class TestUnparseWithMockIntoBasicTypes(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.parser = RowParser(MainModel, MockCellParser())

    def test_input_1(self):
        output1 = self.parser.unparse_row(input1)
        self.assertEqual(output1, output1_exp)

    def test_input_2(self):
        output2 = self.parser.unparse_row(input2)
        self.assertEqual(output2, output2_exp)

    def test_input_3(self):
        output3 = self.parser.unparse_row(input3)
        self.assertEqual(output3, output3_exp)

    def test_input_4(self):
        output4 = self.parser.unparse_row(input4)
        self.assertEqual(output4, output4_exp)

    def test_remap(self):
        output_remap = self.parser.unparse_row(input_remap)
        self.assertEqual(output_remap, output_remap_exp)

    def test_clashingremap1(self):
        output_remap = self.parser.unparse_row(input_clashing_remap1)
        self.assertEqual(output_remap, output_clashing_remap_exp1)

    def test_clashingremap2(self):
        output_remap = self.parser.unparse_row(input_clashing_remap2)
        self.assertEqual(output_remap, output_clashing_remap_exp2)

    def test_clashingremap3(self):
        with self.assertRaises(RowParserError):
            self.parser.unparse_row(input_clashing_remap3)


class TestToNestedList(unittest.TestCase):
    def setUp(self):
        self.parser = RowParser(MainModel, MockCellParser())

    def compare_to_nested_list(self, inp, outp_exp):
        outp = self.parser.to_nested_list(inp)
        self.assertEqual(outp, outp_exp)

    def test_to_nested_list(self):
        self.compare_to_nested_list("abc", "abc")
        self.compare_to_nested_list(["a"], ["a"])
        self.compare_to_nested_list(["a", "c"], ["a", "c"])
        self.compare_to_nested_list(["a", ["c"]], ["a", ["c"]])
        in1 = ModelWithStuff(**submodel_2_dict)
        out1 = [
            ["list_str", ["a", "b", "c"]],
            ["int_field", 5],
            ["str_field", "string"],
        ]
        self.compare_to_nested_list(in1, out1)
        self.compare_to_nested_list([in1], [out1])
        output2 = [
            ["str_field", "main model string"],
            ["model_default", out1],
        ]
        self.compare_to_nested_list(input2, output2)


class ModelWithBasicFields(ParserModel):
    int_field: int = 0
    str_field: str = ""


class MetaModel(ParserModel):
    basic_model: ModelWithBasicFields = ModelWithBasicFields()
    string: str = ""


class MetaModelList(ParserModel):
    basic_model_list: List[ModelWithBasicFields] = []
    model_with_stuff: ModelWithStuff = ModelWithStuff()


class TestUnparseToStringDict(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.parser = RowParser(MainModel, CellParser())

        self.mws = ModelWithStuff(**submodel_2_dict)

        self.metainstance = MetaModel(
            basic_model=ModelWithBasicFields(int_field=42, str_field="word"),
            string="metaword",
        )

        self.metalistinstance = MetaModelList(
            basic_model_list=[
                ModelWithBasicFields(int_field=42, str_field="word"),
                ModelWithBasicFields(int_field=14, str_field="draw"),
            ],
            model_with_stuff=self.mws,
        )

    def test_submodel(self):
        output1 = self.parser.unparse_row(self.mws)
        exp1 = {
            "list_str.1": "a",
            "list_str.2": "b",
            "list_str.3": "c",
            "int_field": 5,
            "str_field": "string",
        }
        self.assertEqual(output1, exp1)

        output2 = self.parser.unparse_row(self.mws, target_headers={"list_str"})
        exp2 = {"list_str": "a|b|c", "int_field": 5, "str_field": "string"}
        self.assertEqual(output2, exp2)

    def test_metamodel(self):
        output1 = self.parser.unparse_row(self.metainstance)
        exp1 = {
            "basic_model.int_field": 42,
            "basic_model.str_field": "word",
            "string": "metaword",
        }
        self.assertEqual(output1, exp1)

        output2 = self.parser.unparse_row(
            self.metainstance, target_headers={"basic_model"}
        )
        exp2 = {
            "basic_model": "int_field;42|str_field;word",
            "string": "metaword",
        }
        self.assertEqual(output2, exp2)

    def test_metamodellist(self):
        output1 = self.parser.unparse_row(self.metalistinstance)
        exp1 = {
            "basic_model_list.1.int_field": 42,
            "basic_model_list.1.str_field": "word",
            "basic_model_list.2.int_field": 14,
            "basic_model_list.2.str_field": "draw",
            "model_with_stuff.list_str.1": "a",
            "model_with_stuff.list_str.2": "b",
            "model_with_stuff.list_str.3": "c",
            "model_with_stuff.int_field": 5,
            "model_with_stuff.str_field": "string",
        }
        self.assertEqual(output1, exp1)

        output2 = self.parser.unparse_row(
            self.metalistinstance,
            target_headers={"model_with_stuff.list_str", "basic_model_list.1"},
        )
        exp2 = {
            "basic_model_list.1": "int_field;42|str_field;word",
            "basic_model_list.2.int_field": 14,
            "basic_model_list.2.str_field": "draw",
            "model_with_stuff.list_str": "a|b|c",
            "model_with_stuff.int_field": 5,
            "model_with_stuff.str_field": "string",
        }
        self.assertEqual(output2, exp2)

    def test_asterisk1(self):
        output2 = self.parser.unparse_row(self.mws, target_headers={"*"})
        exp2 = {"list_str": "a|b|c", "int_field": 5, "str_field": "string"}
        self.assertEqual(output2, exp2)

    def test_asterisk2(self):
        output2 = self.parser.unparse_row(
            self.metalistinstance,
            target_headers={"model_with_stuff.*", "basic_model_list.*"},
        )
        exp2 = {
            "basic_model_list.1": "int_field;42|str_field;word",
            "basic_model_list.2": "int_field;14|str_field;draw",
            "model_with_stuff.list_str": "a|b|c",
            "model_with_stuff.int_field": 5,
            "model_with_stuff.str_field": "string",
        }
        self.assertEqual(output2, exp2)

    def test_exclude(self):
        output1 = self.parser.unparse_row(
            self.metalistinstance,
            excluded_headers={"model_with_stuff.int_field", "basic_model_list.1"},
        )
        exp1 = {
            "basic_model_list.2.int_field": 14,
            "basic_model_list.2.str_field": "draw",
            "model_with_stuff.list_str.1": "a",
            "model_with_stuff.list_str.2": "b",
            "model_with_stuff.list_str.3": "c",
            "model_with_stuff.str_field": "string",
        }
        self.assertEqual(output1, exp1)



from unittest import TestCase

import tablib
from rpft.parsers.common.sheetparser import SheetParser

from tests.mocks import MockRowParser


class TestSheetParser(TestCase):
    def setUp(self):
        self.parser = SheetParser(
            tablib.Dataset(
                ("row1f1", "row1f2"),
                ("row2f1", "row2f2"),
                ("row3f1", "row3f2"),
                headers=("field1", "field2"),
            ),
            row_parser=MockRowParser(),
        )

    def test_context_and_bookmarks(self):
        self.assertEqual(
            self.parser.parse_next_row(),
            {
                "field1": "row1f1",
                "field2": "row1f2",
                "context": {},
            },
        )

        self.parser.create_bookmark("row2")
        self.parser.add_to_context("key", "value")
        self.assertEqual(
            self.parser.parse_next_row(),
            {
                "field1": "row2f1",
                "field2": "row2f2",
                "context": {"key": "value"},
            },
        )

        self.parser.remove_from_context("key")
        self.assertEqual(
            self.parser.parse_next_row(),
            {
                "field1": "row3f1",
                "field2": "row3f2",
                "context": {},
            },
        )

        self.parser.go_to_bookmark("row2")
        self.assertEqual(
            self.parser.parse_next_row(),
            {
                "field1": "row2f1",
                "field2": "row2f2",
                "context": {},
            },
        )

    def test_parse_all(self):
        rows = self.parser.parse_all()

        self.assertEqual(len(rows), 3)
        self.assertEqual(
            rows[0],
            {
                "field1": "row1f1",
                "field2": "row1f2",
                "context": {},
            },
        )
        self.assertEqual(
            rows[1],
            {
                "field1": "row2f1",
                "field2": "row2f2",
                "context": {},
            },
        )
        self.assertEqual(
            rows[2],
            {
                "field1": "row3f1",
                "field2": "row3f2",
                "context": {},
            },
        )



from unittest import TestCase
from typing import List

from rpft.parsers.common.cellparser import CellParser, unescape
from rpft.parsers.common.rowparser import ParserModel


class InnerModel(ParserModel):
    str_field: str


class OuterModel(ParserModel):
    inner: InnerModel
    strings: List[str]


class TestStringSplitter(TestCase):
    def setUp(self):
        self.parser = CellParser()

    def compare_split_by_separator(self, string, exp):
        out = self.parser.split_by_separator(string, "|")
        self.assertEqual(out, exp)

    def compare_split_into_lists(self, string, exp):
        out = self.parser.split_into_lists(string)
        self.assertEqual(out, exp)

    def compare_join_from_lists(self, string, exp):
        out = self.parser.join_from_lists(string)
        self.assertEqual(out, exp)

    def test_split_by_separator(self):
        self.compare_split_by_separator("a", "a")
        self.compare_split_by_separator("a|b", ["a", "b"])
        self.compare_split_by_separator("a|b|", ["a", "b"])
        self.compare_split_by_separator("a|", ["a"])
        self.compare_split_by_separator("a\\|", "a\\|")

    def test_unescape(self):
        tests = [
            ("a", "a", "Strings without escape sequence should not change"),
            (r"a\|", "a|", "Escape sequence should be removed"),
            (r"a\;", "a;", "Escape sequence should be removed"),
            (r"a\\", "a\\", "Escape sequence should be removed"),
            (r"a\\;", r"a\;", "Escape sequence should be removed"),
            ([[[r"a\;"]]], [[["a;"]]], "List items should be recursively cleansed"),
            ([r"\\", r"\;"], ["\\", ";"], "List items should be recursively cleansed"),
            (" a;\n", "a;", "White space must be removed"),
            ([" a;\n"], ["a;"], "White space must be removed"),
        ]

        for data, expected, msg in tests:
            self.assertEqual(unescape(data), expected, msg)

    SPLIT_TESTS = {
        # Not part of rejoin tests because of non-toplevel separator
        # or unneccesary trailing separator
        "1;": ["1"],
        "1;2": ["1", "2"],
        "1;2;": ["1", "2"],
        "1;2;;": ["1", "2", ""],
        "a|b|": ["a", "b"],
        "1;2|3;4;": [["1", "2"], ["3", "4"]],
    }

    REJOIN_TESTS = {
        # Not part of splitter tests because of non-string types
        "0": 0,
        "True": True,
        "1.0": 1.0,
        "0|True|1.0": [0, True, 1.0],
    }

    SPLIT_REJOIN_TESTS = {
        "1": "1",
        "a\\|": "a|",
        "a\\;": "a;",
        "abc": "abc",
        "a|b|c": ["a", "b", "c"],
        "a|b;c": ["a", ["b", "c"]],
        "a;|": [["a"]],
        "1;2|": [["1", "2"]],
        "1;|2;": [["1"], ["2"]],
        "1;|2": [["1"], "2"],
        "|a": ["", "a"],
        "a|b": ["a", "b"],
        "a|": ["a"],
        "a|\\;": ["a", ";"],
        "a|;": ["a", [""]],
        "1;2|3;4": [["1", "2"], ["3", "4"]],
        "1;2|3;4\\;": [["1", "2"], ["3", "4;"]],
        "1;2|3;4\\|": [["1", "2"], ["3", "4|"]],
        CellParser.escape_string("|;;|\\;\\\\\\|"): "|;;|\\;\\\\\\|",
    }

    def test_split_into_lists(self):
        for inp, outp in TestStringSplitter.SPLIT_TESTS.items():
            self.compare_split_into_lists(inp, outp)
        for inp, outp in TestStringSplitter.SPLIT_REJOIN_TESTS.items():
            self.compare_split_into_lists(inp, outp)

    def test_join_from_lists(self):
        for outp, inp in TestStringSplitter.REJOIN_TESTS.items():
            self.compare_join_from_lists(inp, outp)
        for outp, inp in TestStringSplitter.SPLIT_REJOIN_TESTS.items():
            self.compare_join_from_lists(inp, outp)

    def test_string_stripping(self):
        self.compare_split_into_lists(" a\n|\nb ", ["a", "b"])
        self.compare_split_into_lists("1; 2\n|\n3; 4", [["1", "2"], ["3", "4"]])


class TestCellParser(TestCase):
    def setUp(self):
        self.parser = CellParser()

    def test_strings_without_templates_are_not_changed(self):
        self.assertEqual(
            CellParser().parse_as_string("plain string"),
            ("plain string", False),
        )

    def test_templates_are_rendered(self):
        self.assertEqual(
            CellParser().parse_as_string(" {{var}} string ", context={"var": "abc"}),
            ("abc string", False),
        )

    def test_parse(self):
        out = self.parser.parse("a;b;c")
        self.assertEqual(out, ["a", "b", "c"])

        context = {"list": ["a", "b", "c"]}
        out = self.parser.parse("{% for e in list %}{{e}}{% endfor %}", context=context)
        self.assertEqual(out, "abc")
        # Templating comes first, only then splitting into a list
        out = self.parser.parse(
            "{% for e in list %}{{e}};{% endfor %}", context=context
        )
        self.assertEqual(out, ["a", "b", "c"])

    def test_escape_filter(self):
        out = self.parser.parse('{{"a;b;c"}}')
        self.assertEqual(out, ["a", "b", "c"])
        out = self.parser.parse('{{"a;b;c"|escape}}')
        self.assertEqual(out, "a;b;c")
        string = "\\|;\\"
        out = self.parser.parse("{{string|escape}}", context={"string": string})
        self.assertEqual(out, string)

    def test_indicate_is_object_after_parsing_native_type_template(self):
        self.assertEqual(
            self.parser.parse_as_string(
                ' {@(1, 2, [input == "yes", "a"])@} ',
                context={"input": "yes"},
            ),
            ((1, 2, [True, "a"]), True),
            "Rendered value should not be string; is_object should be True",
        )
